// File: InputChoice.java

package kero.awt;

import java.util.*;
import java.awt.*;
import java.awt.event.*;

/**
 * Combines a Label and a Choice into one Component, such that it functions
 * as an input choice. When an ActionEvent from the Choice is caught, the
 * event is passed onto objects that registered as ActionListener to an
 * instance of this class.
 *
 * <P>The constuctions for passing events from the choice to the listeners
 * of the entire InputChoice are bad; they should use the mechanism provided by
 * the jdk. Then the for-loop can be synchronized. Only then the events are
 * completely thread-safe.
 *
 * @author    Kero van Gelder
 * @copyright Kero van Gelder (the GNU General Library Public License applies)
 * @since     8 I 2000
 * @version   9 I 2000
 */
public class InputChoice
  extends Panel
  implements ItemListener, KeyListener, FocusListener
{
  Label label;
  Choice choice;
  Vector action, focus, key;

  public InputChoice(String labeltext) {
      super(new FlowLayout(1, 0, 0));
    label = new Label(labeltext);
    super.add(label);
    choice = new Choice();
    choice.addItemListener(this);
    choice.addKeyListener(this);
    choice.addFocusListener(this);
    super.add(choice);
    action = new Vector();
    focus = new Vector();
    key = new Vector();
  }


  // admission and withdrawal of several event listeners

  public void addActionListener(ActionListener listener) {
    action.addElement(listener);
  }

  public void removeActionListener(ActionListener listener) {
    action.removeElement(listener);
  }

  public void addFocusListener(FocusListener listener) {
    focus.addElement(listener);
  }

  public void removeFocusListener(FocusListener listener) {
    focus.removeElement(listener);
  }

  public void addKeyListener(KeyListener listener) {
    key.addElement(listener);
  }

  public void removeKeyListener(KeyListener listener) {
    key.removeElement(listener);
  }


  public void itemStateChanged(ItemEvent ie) {
    throwActionPerformed((String)ie.getItem());
  }

  // one ActionListener method

  public void actionPerformed(ActionEvent ae) {
    throwActionPerformed(ae.getActionCommand());
  }

  protected void throwActionPerformed(String text) {
    ActionEvent event;
    for (int i=0; i<action.size(); i++) {
      event = new ActionEvent(this, ActionEvent.ACTION_PERFORMED, text);
      ((ActionListener)action.elementAt(i)).actionPerformed(event);
    }
  }


  // three KeyListener methods

  public void keyPressed(KeyEvent ke) {
    // System.err.println("InputChoice.keyPressed() "+ke);
    switch (ke.getKeyCode()) {
    case KeyEvent.VK_ESCAPE:
      // cancel editing
      // choice.closeSelectionMechanism(); // darn, does not exist!!!
      break;
    default:
      // ignore
      break;
    }
    for (int i=0; i<key.size(); i++) {
      KeyEvent event;
      event = new KeyEvent(this, ke.getID(), ke.getWhen(),
			   ke.getModifiers(), ke.getKeyCode(),
			   ke.getKeyChar());
      ((KeyListener)key.elementAt(i)).keyPressed(event);
    }
  }

  public void keyReleased(KeyEvent ke) {
    for (int i=0; i<key.size(); i++) {
      KeyEvent event;
      event = new KeyEvent(this, ke.getID(), ke.getWhen(),
			   ke.getModifiers(), ke.getKeyCode(),
			   ke.getKeyChar());
      ((KeyListener)key.elementAt(i)).keyReleased(event);
    }
  }

  public void keyTyped(KeyEvent ke) {
    for (int i=0; i<key.size(); i++) {
      KeyEvent event;
      event = new KeyEvent(this, ke.getID(), ke.getWhen(),
			   ke.getModifiers(), ke.getKeyCode(),
			   ke.getKeyChar());
      ((KeyListener)key.elementAt(i)).keyTyped(event);
    }
  }


  // two FocusListener methods plus additional Focus method

  public void focusGained(FocusEvent fe) {
    // ignore for self
    // pass on to subscribed FocusListeners
    for (int i=0; i<focus.size(); i++) {
      FocusEvent event;
      event = new FocusEvent(this, FocusEvent.FOCUS_GAINED, fe.isTemporary());
      ((FocusListener)focus.elementAt(i)).focusGained(event);
    }
  }

  public void focusLost(FocusEvent fe) {
    // set value for self
    //if (config.setOnFocusLost())
    // pass on to subscribed FocusListeners
    for (int i=0; i<focus.size(); i++) {
      FocusEvent event;
      event = new FocusEvent(this, FocusEvent.FOCUS_LOST, fe.isTemporary());
      ((FocusListener)focus.elementAt(i)).focusLost(event);
    }
    throwActionPerformed(choice.getSelectedItem());
  }

  public void requestFocus() {
    choice.requestFocus();
  }


  // Normal interaction methods

  public void setLabel(String labeltext) {
    label.setText(labeltext);
  }

  public String getLabel() {
    return label.getText();
  }

  public void setLabelFont(Font font) {
    label.setFont(font);
  }

  /**
   * Preferably do not use this method. It may disappear in future versions.
   */
  public Choice getChoice() {
    return choice;
  }

  public String getSelection() {
    return choice.getSelectedItem();
  }

  public int getIndex() {
    return choice.getSelectedIndex();
  }

  public void setSelection(String text) {
    choice.select(text);
  }

  public void setSelection(int index) {
    choice.select(index);
  }

  public void setTextFont(Font font) {
    choice.setFont(font);
  }

  public void addSelection(String text) {
    choice.addItem(text);
  }

  public void removeAll() {
    choice.removeAll();
  }

  // overridden Container methods

  public final Component add(Component component) {
    System.err.println("An InputChoice is not a Container as such.\n"+
		       "Component not added.\n");
    return component;
  }

  public Dimension getPreferredSize() {
    Dimension labelsize = label.getPreferredSize();
    Dimension textsize = choice.getPreferredSize();
    int prefwidth = labelsize.width + textsize.width;
    int prefheight = labelsize.height;
    if (textsize.height > prefheight) prefheight = textsize.height;
    return new Dimension(prefwidth, prefheight);
  }

  public void doLayout() {
    //System.err.println("doLayout()\n"+
    //		       "size: "+getSize()+"\n"+
    //	               "  insets: "+getInsets()+"\n");
    Dimension ownsize=getSize();
    Dimension labelsize=label.getPreferredSize();
    Dimension textsize=choice.getPreferredSize();
    label.setSize(labelsize.width, textsize.height);
    label.setLocation(0,0);
    choice.setSize(ownsize.width-labelsize.width, textsize.height);
    choice.setLocation(labelsize.width,0);
  }
}
