// File: InputField.java

package kero.awt;

import java.util.*;
import java.awt.*;
import java.awt.event.*;

/**
 * Combines a Label and a TextField into one Component, such that it functions
 * as an input field. When an ActionEvent from the TextField is caught, the
 * event is passed onto objects that registered as ActionListener to an
 * instance of this class.
 *
 * <P>The constuctions for passing events from the textfield to the listeners
 * of the entire InputField are bad; they should use the mechanism provided by
 * the jdk. Then the for-loop can be synchronized. Only then the events are
 * completely thread-safe.
 *
 * @author    Kero van Gelder
 * @copyright Kero van Gelder (the GNU General Library Public License applies)
 * @since     11 IX 1998
 * @version   23 I 2000
 */
public class InputField
  extends Panel
  implements ActionListener, KeyListener, FocusListener
{
  Label label;
  TextField textfield;
  String value;
  Vector action, focus, key;
  boolean usefocus;

  public InputField(String labeltext, String fieldtext) {
    this(labeltext, fieldtext, true);
  }

  public InputField(String labeltext, String fieldtext, boolean usefocus) {
    super(new FlowLayout(1,0,0));
    label = new Label(labeltext);
    super.add(label);
    textfield = new TextField(fieldtext);
    textfield.addFocusListener(this);
    super.add(textfield);
    value = fieldtext;
    this.usefocus = usefocus;
    action = new Vector();
    focus = new Vector();
    key = new Vector();
    activate();
  }


  // do I still need these two...

  public void deactivate() {
    textfield.setEditable(false);
    //textfield.disable();
    textfield.removeActionListener(this);
    textfield.removeKeyListener(this);
  }

  public void activate() {
    textfield.setEditable(true);
    //textfield.enable();
    textfield.addActionListener(this);
    textfield.addKeyListener(this);
  }


  // admission and withdrawal of several event listeners

  public void addActionListener(ActionListener listener) {
    action.addElement(listener);
  }

  public void removeActionListener(ActionListener listener) {
    action.removeElement(listener);
  }

  public void addMouseListener(MouseListener listener) {
    textfield.addMouseListener(listener);
  }

  public void removeMouseListener(MouseListener listener) {
    textfield.removeMouseListener(listener);
  }

  public void addFocusListener(FocusListener listener) {
    focus.addElement(listener);
  }

  public void removeFocusListener(FocusListener listener) {
    focus.removeElement(listener);
  }

  public void addKeyListener(KeyListener listener) {
    key.addElement(listener);
  }

  public void removeKeyListener(KeyListener listener) {
    key.removeElement(listener);
  }


  // one ActionListener method

  public void actionPerformed(ActionEvent ae) {
    value = ae.getActionCommand();
    throwActionPerformed(ae);
  }

  protected void throwActionPerformed(ActionEvent ae) {
    String text = "";
    if (ae != null) text = ae.getActionCommand();
    ActionEvent event;
    for (int i=0; i<action.size(); i++) {
      event = new ActionEvent(this, ActionEvent.ACTION_PERFORMED, text);
      ((ActionListener)action.elementAt(i)).actionPerformed(event);
    }
  }


  // three KeyListener methods

  public void keyPressed(KeyEvent ke) {
    // System.err.println("InputField.keyPressed() "+ke);
    switch (ke.getKeyCode()) {
    case KeyEvent.VK_ESCAPE:
      // cancel editing
      textfield.setText(value);
      break;
    default:
      // ignore
      break;
    }
    for (int i=0; i<key.size(); i++) {
      KeyEvent event;
      event = new KeyEvent(this, ke.getID(), ke.getWhen(),
			   ke.getModifiers(), ke.getKeyCode(),
			   ke.getKeyChar());
      ((KeyListener)key.elementAt(i)).keyPressed(event);
    }
  }

  public void keyReleased(KeyEvent ke) {
    for (int i=0; i<key.size(); i++) {
      KeyEvent event;
      event = new KeyEvent(this, ke.getID(), ke.getWhen(),
			   ke.getModifiers(), ke.getKeyCode(),
			   ke.getKeyChar());
      ((KeyListener)key.elementAt(i)).keyReleased(event);
    }
  }

  public void keyTyped(KeyEvent ke) {
    for (int i=0; i<key.size(); i++) {
      KeyEvent event;
      event = new KeyEvent(this, ke.getID(), ke.getWhen(),
			   ke.getModifiers(), ke.getKeyCode(),
			   ke.getKeyChar());
      ((KeyListener)key.elementAt(i)).keyTyped(event);
    }
  }


  // two FocusListener methods plus additional Focus method

  public void focusGained(FocusEvent fe) {
    // ignore for self
    // pass on to subscribed FocusListeners
    for (int i=0; i<focus.size(); i++) {
      FocusEvent event;
      event = new FocusEvent(this, FocusEvent.FOCUS_GAINED, fe.isTemporary());
      ((FocusListener)focus.elementAt(i)).focusGained(event);
    }
  }

  public void focusLost(FocusEvent fe) {
    // set value for self
    //if (config.setOnFocusLost())
    value = textfield.getText();
    // pass on to subscribed FocusListeners
    for (int i=0; i<focus.size(); i++) {
      FocusEvent event;
      event = new FocusEvent(this, FocusEvent.FOCUS_LOST, fe.isTemporary());
      ((FocusListener)focus.elementAt(i)).focusLost(event);
    }
    if (usefocus) {
      throwActionPerformed(
        new ActionEvent(this, ActionEvent.ACTION_PERFORMED, value)
      );
    }
  }

  public void requestFocus() {
    textfield.requestFocus();
  }


  // Normal interaction methods

  public void setLabel(String labeltext) {
    label.setText(labeltext);
  }

  public String getLabel() {
    return label.getText();
  }

  public void setLabelFont(Font font) {
    label.setFont(font);
  }

  /**
   * Preferably do not use this method. It may disappear in future versions.
   */
  public TextField getTextField() {
    return textfield;
  }

  public String getText() {
    return textfield.getText();
  }

  public void setText(String text) {
    textfield.setText(text);
  }

  public void setTextFont(Font font) {
    textfield.setFont(font);
  }

  // the (il)legal part does not function well !!!
  
  /**
   * Internal reverter of setIllegal()
   */
  protected void setLegal() {
    textfield.setForeground(Color.black);
  }

  /**
   * Meant to be called by inputFieldEvent-receiver when input-value was
   * illegal. Will be unset when new input has been given (needs to be tested
   * with a test-run...).
   */
  public void setIllegal() {
    textfield.selectAll();
    textfield.setForeground(Color.red);
  }


  // overridden Container methods

  public final Component add(Component component) {
    System.err.println("An InputField is not a Container as such.\n"+
		       "Component not added.\n");
    return component;
  }

  public Dimension getPreferredSize() {
    Dimension labelsize = label.getPreferredSize();
    Dimension textsize = textfield.getPreferredSize();
    int prefwidth = labelsize.width + textsize.width;
    int prefheight = labelsize.height;
    if (textsize.height > prefheight) prefheight = textsize.height;
    return new Dimension(prefwidth, prefheight);
  }

  public void dontLayout() {
    //System.err.println("doLayout()\n"+
    //		       "size: "+getSize()+"\n"+
    //	               "  insets: "+getInsets()+"\n");
    Dimension ownsize=getSize();
    Dimension labelsize=label.getPreferredSize();
    Dimension textsize=textfield.getPreferredSize();
    label.setSize(labelsize.width, textsize.height);
    label.setLocation(0,0);
    textfield.setSize(ownsize.width-labelsize.width, textsize.height);
    textfield.setLocation(labelsize.width,0);
  }
}
