// File: MyContainer.java

package kero.awt;

import java.lang.reflect.*;
import java.util.Hashtable;
import java.awt.*;
import java.awt.event.*;
import kero.util.*;
import kero.awt.event.InputEvent;
import kero.awt.event.InputListener;
import kero.awt.event.InputCollector;

/**
 * Supposed to catch all InputEvents and deal with them in a uniform way.
 *
 * <P>Replaces MyFrame as such.
 *
 * <P>Since multiple inheritance is not implemented in Java and I want to
 * implement this exactly once (hey, OO design need not be nice...) I need to
 * do it in a sensible place. A Container has specific (bound!) functionality.
 * Though a Frame should do the same, a MyContainer can be put in a Frame and
 * catch all events for the Frame.
 *
 * <P>An advantage of catching events in a Container, is that the same
 * keystroke will be caught for each component and can do the same thing,
 * whereever it came from. (When different things need to be done, you are not
 * in a worse situation than without MyContainer.)
 *
 * @author Kero van Gelder
 * @copyright Kero van Gelder (the GNU General Public License applies)
 * @since 5 I 2001
 * @version 13 I 2001
 */
public class MyContainer
    extends Container
    implements InputListener, Reportable, KeyListener
{
    InputCollector collector;
    Hashtable source2methodName, key2focus;

    public MyContainer() {
	super();
	collector = new InputCollector();
	collector.addInputListener(this);
	source2methodName = new Hashtable();
	key2focus = new Hashtable();
    }

    /**
     * Adds the Component to this and passes the keyCode-KeyEvent (and all
     * other applicable events of the Component) to this.
     */
    public Component add(Component comp,
			 int keyCode, String method) {
	Component result = super.add(comp);
	setInputCollector(comp, keyCode, method);
	return result;
    }

    public Component add(Component comp, int index,
			 int keyCode, String method) {
	Component result = super.add(comp, index);
	setInputCollector(comp, keyCode, method);
	return result;
    }

    public void add(Component comp, Object constraints,
		    int keyCode, String method) {
	super.add(comp, constraints);
	setInputCollector(comp, keyCode, method);
    }

    public void add(Component comp, Object constraints, int index,
		    int keyCode, String method) {
	super.add(comp, constraints, index);
	setInputCollector(comp, keyCode, method);
    }

    /**
     * Like add(), but put both the String (as Label) and the Component to
     * this, joined together in a Panel.
     * @return The constructed Panel.
     */
    public Component add(String s, Component comp,
			 int keyCode, String method) {
	Panel p = new InputComponent(s, comp);
	super.add(p);
	setInputCollector(comp, keyCode, method);
	return p;
    }

    public Component add(String s, Component comp, int index,
			 int keyCode, String method) {
	Panel p = new InputComponent(s, comp);
	super.add(p, index);
	setInputCollector(comp, keyCode, method);
	return p;
    }

    public void add(String s, Component comp, Object constraints,
		    int keyCode, String method) {
	Panel p = new InputComponent(s, comp);
	super.add(p, constraints);
	setInputCollector(comp, keyCode, method);
    }

    public void add(String s, Component comp, Object constraints, int index,
		    int keyCode, String method) {
	Panel p = new InputComponent(s, comp);
	super.add(p, constraints, index);
	setInputCollector(comp, keyCode, method);
    }


    /**
     * Like add(), but puts the Component in the Container, not in this. The
     * Events are still passed to this for convenient grouping of events.
     */
    public Component addTo(Container cont,
			   Component comp,
			   int keyCode, String method) {
	Component result = cont.add(comp);
	setInputCollector(comp, keyCode, method);
	return result;
    }

    public Component addTo(Container cont,
			   Component comp, int index,
			   int keyCode, String method) {
	Component result = cont.add(comp, index);
	setInputCollector(comp, keyCode, method);
	return result;
    }

    public void addTo(Container cont,
		      Component comp, Object constraints,
		      int keyCode, String method) {
	cont.add(comp, constraints);
	setInputCollector(comp, keyCode, method);
    }

    public void addTo(Container cont, 
		      Component comp, Object constraints, int index,
		      int keyCode, String method) {
	cont.add(comp, constraints, index);
	setInputCollector(comp, keyCode, method);
    }


    // Stuff to deal with advanced add()ing

    private void setInputCollector(Component comp,
				   int keyCode, String methodName) {
	Component c = comp;
	if (c instanceof Button) {
	    ((Button)c).addActionListener(collector);
	} else if (c instanceof TextField) {
	    ((TextField)c).addActionListener(collector);
	    ((TextField)c).addFocusListener(collector);
	} else if (c instanceof Choice) {
	    ((Choice)c).addItemListener(collector);
	}
	c.addKeyListener(this);
	source2methodName.put(c, methodName);
	key2focus.put(new Integer(keyCode), comp);
	if (keyCode != 0) {
	    Object o = source2methodName.put(new Integer(keyCode), methodName);
	    if (o != null) warn("removed "+o+" from source2methodName");
	}
    }

    public void keyPressed(KeyEvent ke) {
	/* if a modified keycode was given... */
	int altHeld = ke.getModifiers() & KeyEvent.ALT_MASK;
	int ctrlHeld = ke.getModifiers() & KeyEvent.CTRL_MASK;
	if ((altHeld == 0) && (ctrlHeld == 0)) return;

	/* ...we want to either activate the associated Component (when the
	 * input itself still needs to be given, e.g. a TextField) or execute
	 * the associated Component (when there is only one possible action,
	 * e.g. a Button).
	 */
	warn("MyContainer.keyPressed() source = "+ke.getSource());
	Integer keyCode = new Integer(ke.getKeyCode());
	Component target = (Component)key2focus.get(keyCode);
	if ((target == null) || (target == ke.getSource())) return;
	// should trigger input event for source, somehow
	if (target instanceof Button) {
	    inputGiven(new InputEvent(target, InputEvent.INPUT_GIVEN));
	} else {
	    target.requestFocus();
	}
    }

    public void keyReleased(KeyEvent ke) {}
    public void keyTyped(KeyEvent ke) {}


    // InputListener method

    public void inputGiven(InputEvent ie) {
	String methodName = (String)source2methodName.get(ie.getSource());
	warn("methodName = "+methodName);
	if ((methodName == null) || ("".equals(methodName))) return;
	Method method[] = getClass().getMethods();
	for (int i=0; i<method.length; i++) { // Hashtable would be lovely, Sun
	    if (method[i].getName().equals(methodName)) {
		Class param[] = method[i].getParameterTypes();
		if (param.length != 0) continue;
		try {
		    warn("MyContainer.inputGiven() calls "+method[i].getName());
		    method[i].invoke(this, null);
		} catch (IllegalAccessException iaccess) {
		    crash(255, "MyFrame.invoke() "+iaccess);
		} catch (IllegalArgumentException iarge) {
		    crash(255, "MyFrame.invoke() "+iarge);
		} catch (InvocationTargetException ite) {
		    crash(255, "MyFrame.invoke() "+ite);
		}
		return;
	    }
	}
	crash(255, "Method not implemented, sorry");
    }


    // Reportable methods

    public void setReportable(Reportable r) {
	// ignore (should be implemented sometime)
    }

    public void unsetReportable() {
	// ignore (should be implemented when setReportable() is implemented)
    }

    public void report(String message) {
	System.err.println(message);
    }

    public void warn(String message) {
	System.err.println(message);
    }

    public void crash(int errno, String message) {
	System.err.println(message);
	System.exit(errno);
    }
}
