// File: MyFrame.java

package kero.awt;

import java.util.*;
import java.lang.reflect.*;
import java.awt.*;
import java.awt.event.*;
import kero.util.*;

/**
 * Is a Frame with a lot of extra functionality:
 * <UL>
 *  <LI>Terminates the JVM when all <CODE>MyFrame</code>s are closed (together
 *      with <CODE>FrameAdapter</code>),
 *  <LI>Coupling and automatic calling of methods for <CODE>Button</code>s and
 *      some other java.awt.Components.
 * </ul>
 *
 * @see java.awt.Frame
 * @see java.awt.event.WindowListener
 * @see kero.awt.FrameAdapter
 *
 * @author    Kero van Gelder
 * @copyright Kero van Gelder (the GNU General Library Public License applies)
 * @since     6 IX 1998
 * @version   24 II 2001
 */
public class MyFrame
    extends Frame
    implements ActionListener, KeyListener, Reportable
{
    Hashtable source2methodName, source2focus;
    boolean eventsDisabled = false;
    protected Reportable reportable = null;

    public MyFrame() {
	this("");
    }

    public MyFrame(String framename) {
	super(framename);
	source2methodName = new Hashtable();
	source2focus = new Hashtable();
    }

  public void limitSize() {
    Dimension screenSize = getToolkit().getScreenSize();
    Dimension preferredSize = getPreferredSize();
    int desWidth = preferredSize.width;
    int desHeight = preferredSize.height;
    if (screenSize.width - 40 < desWidth) desWidth = screenSize.width - 40;
    if (screenSize.height - 40 < desHeight) desHeight = screenSize.height - 40;
    setSize(desWidth, desHeight);
  }

    /**
     * Adds this frame to a listener.
     * @see addWindowListener
     */
    public void addWindowListener(WindowListener wl) {
	if (wl instanceof FrameAdapter) {
	    FrameAdapter fa = (FrameAdapter)wl;
	    if (isShowing()) {
		// Window is already visible, hence the windowOpened() event
		// will never arrive. Either post a WindowOpenedEvent or raise
		// framecount by hand:
		fa.setFrameCount(fa.getFrameCount()+1); // effect of the event
	    }
	}
	super.addWindowListener(wl);
    }

    /**
     * Called when MyFrame is closed. Can be overriden excellently (read: do so
     * by all means).
     */
    public void closeFrame() {
	dispose();
    }


    // methods for adding and listening to Components, plus invoking methods

    /**
     * Associates a method with both a Button and a KeyCode, that is, the
     * method will be called when either the Button is pressed or the KeyCode
     * is given to a Component that has this MyFrame as Listener.
     * <BR>The method will be called without any parameters.
     */
    public void addButton(Button button,
			  int keyCode, String methodName) {
	addButtonTo(this, button, keyCode, methodName);
    }

    /**
     * Associates a method with both a Button and a KeyCode, that is, the
     * method will be called when either the Button is pressed or the KeyCode
     * is given to a Component that has this MyFrame as Listener.
     * <BR>The Button will be put in the Container, not in MyFrame (though
     * typical usage has the Container as (grand)child of MyFrame).
     * <BR>The method will be called without any parameters.
     */
    public void addButtonTo(Container container, Button button,
			    int keyCode, String methodName) {
	button.addActionListener(this);
	button.addKeyListener(this);
	container.add(button);
	source2methodName.put(button, methodName);
	source2methodName.put(new Integer(keyCode), methodName);
	source2focus.put(new Integer(keyCode), button);
    }

    public void addInput(InputChoice choice,
			 int keyCode, String methodName) {
	addInputTo(this, choice, keyCode, methodName);
    }

    public void addInputTo(Container container, InputChoice choice,
			   int keyCode, String methodName) {
	choice.addActionListener(this);
	choice.addKeyListener(this);
	container.add(choice);
	source2methodName.put(choice, methodName);
	source2focus.put(new Integer(keyCode), choice);
    }

    public void addInputField(InputField field, int keyCode,
			      String methodName) {
	addInputFieldTo(this, field, keyCode, methodName);
    }

    public void addInputFieldTo(Container container, InputField field,
				int keyCode, String methodName) {
	field.addActionListener(this);
	field.addKeyListener(this);
	container.add(field);
	source2methodName.put(field, methodName);
	source2focus.put(new Integer(keyCode), field);
    }

    public void addNumberInput(NumberInput numberinput, int keyCode,
			       String methodName) {
	addNumberInputTo(this, numberinput, keyCode, methodName);
    }

    public void addNumberInputTo(Container container, NumberInput numberinput,
				 int keyCode, String methodName) {
	numberinput.addActionListener(this);
	numberinput.addKeyListener(this);
	container.add(numberinput);
	source2methodName.put(numberinput, methodName);
	source2methodName.put(new Integer(keyCode), methodName);
	source2focus.put(new Integer(keyCode), numberinput);
    }

    public void disableEvents() {
	eventsDisabled = true;
    }

    public void enableEvents() {
	eventsDisabled = false;
    }

    public void invoke(String methodName) {
	//warn("Want to execute: "+methodName);
	if (eventsDisabled) return;
	if ((methodName == null) || ("".equals(methodName))) return;
	Method method[] = getClass().getMethods();
	for (int i=0; i<method.length; i++) { // Hashtable would be lovely, Sun
	    if (method[i].getName().equals(methodName)) {
		Class param[] = method[i].getParameterTypes();
		if (param.length != 0) continue;
		try {
		    //warn("calling "+method[i].getName());
		    method[i].invoke(this, null);
		} catch (IllegalAccessException iaccess) {
		    crash(255, "MyFrame.invoke() "+iaccess);
		} catch (IllegalArgumentException iarge) {
		    crash(255, "MyFrame.invoke() "+iarge);
		} catch (InvocationTargetException ite) {
		    crash(255, "MyFrame.invoke() "+ite);
		}
		return;
	    }
	}
	crash(255, "Method not implemented, sorry");
    }


    // one ActionListener method

    public void actionPerformed(ActionEvent ae) {
	//warn("XXX.actionPerformed() "+ae);
	Object source = ae.getSource();
	String methodName = (String)source2methodName.get(source);
	invoke(methodName);
    }


    // three KeyListener method

    public void keyTyped(KeyEvent ke) {
	//warn("Key `"+ke.getKeyChar()+"' typed");
    }

    public void keyPressed(KeyEvent ke) {
	//warn("XXX.keyPressed() Got event: "+ke);
	int altHeld = ke.getModifiers() & KeyEvent.ALT_MASK;
	int ctrlHeld = ke.getModifiers() & KeyEvent.CTRL_MASK;
	if ((altHeld == 0) && (ctrlHeld == 0)) return;

	Integer keyCode = new Integer(ke.getKeyCode());
	String methodName = (String)source2methodName.get(keyCode);
	if (methodName != null) invoke(methodName);
	Component target = (Component)source2focus.get(keyCode);
	if (target != null) target.requestFocus();
    }

    public void keyReleased(KeyEvent ke) {
	//warn("Key `"+ke.getKeyChar()+"' released");
    }


    // four Reportable methods

    public void setReportable(Reportable r) {
	reportable = r;
    }

    public void unsetReportable() {
	reportable = null;
    }

    public void report(String message) {
	if (reportable != null) reportable.report(message);
	else System.err.println(message);
    }

    public void warn(String message) {
	if (reportable != null) reportable.warn(message);
	else System.err.println(message);
    }

    public void crash(int errNo, String message) {
	System.err.println(message);
	System.exit(errNo);
    }
}
