// File: NumberInput.java

package kero.awt;

import java.util.*;
import java.awt.*;
import java.awt.event.*;

/**
 * Combines a Label, couple of Buttons and a TextField into one Component, such
 * that it functions as an input field. When an ActionEvent from the TextField
 * is caught, the event is passed onto objects that registered as
 * ActionListener to an instance of this class.
 *
 * <P>The constuctions for passing events from the textfield to the listeners
 * of the entire NumberInput are bad; they should use the mechanism provided by
 * the jdk. Then the for-loop can be synchronized. Only then the events are
 * completely thread-safe.
 *
 * <P>This <B>abstract</b> class is <EM>not</em> meant to hide the details from
 * IntField, DoubleField and the like, but to implement the code in this class
 * only once, instead of for each subclass. An appropriate setNumber() method
 * for each subclass is mandatory.
 *
 * @author    Kero van Gelder
 * @copyright Kero van Gelder (the GNU General Library Public License applies)
 * @since     14 XI 1999
 * @version   19 I 2000
 */
public abstract class NumberInput
  extends Panel
  implements ActionListener, KeyListener, FocusListener
{
  Label label;
  int count; // amount of decrease buttons
  Button dec[], inc[];
  TextField textfield;
  Vector action, focus, key;

  public NumberInput() {
    normalConstructor();
  }

  private void normalConstructor() {
    action = new Vector();
    focus = new Vector();
    key = new Vector();
  }

  /**
   * A subclass calling this ``constructor'' should ensure that the method
   * getNumber() returns something meaningful (that's the reason this is not in
   * the real constructor).
   *
   * <P>This is necessary, because a generic setNumber() can not be made.
   * @param labeltext Text for the label. When <CODE>null</code>, the label is
   * omitted.
   * @param amount Amount of decrease and increase buttons (so actually there
   * is twice the amount).
   */
  protected void constructUI(String labeltext, int amount) {
      setLayout(new InputLayout());
    if (amount < 0) amount = 0;
    if (amount > 3) {
      System.err.println("VERY many buttons here, Mr. Programmer");
    }
    this.count = amount;

    if (labeltext != null) {
      label = new Label(labeltext);
      super.add(label);
    }

    dec = new Button[count];
    String decstr = "";
    for (int i=0; i<count; i++) {
      decstr += "<";
      dec[i] = new Button(decstr);
      dec[i].addActionListener(this);
      dec[i].addKeyListener(this);
    }
    for (int i=count-1; i>=0; i--) super.add(dec[i]);

    textfield = new TextField(getNumber());
    textfield.addFocusListener(this);
    textfield.addKeyListener(this);
    textfield.addActionListener(this);
    super.add(textfield);

    inc = new Button[count];
    String incstr = "";
    for (int i=0; i<count; i++) {
      incstr += ">";
      inc[i] = new Button(incstr);
      inc[i].addActionListener(this);
      inc[i].addKeyListener(this);
      super.add(inc[i]);
    }
  }

  /**
   * return the number in the form of a String (which is the only ``general''
   * form of a number). subclasses are encouraged to write functions like
   * getInt() for IntField, getDouble for DoubleField etc.
   */
  public abstract String getNumber();

  public abstract void setNumberFromField();

  public abstract void lowerNumber(int step);
  public abstract void raiseNumber(int step);

  // admission and withdrawal of several event listeners

  public void addActionListener(ActionListener listener) {
    action.addElement(listener);
  }

  public void removeActionListener(ActionListener listener) {
    action.removeElement(listener);
  }

  public void addMouseListener(MouseListener listener) {
    textfield.addMouseListener(listener);
  }

  public void removeMouseListener(MouseListener listener) {
    textfield.removeMouseListener(listener);
  }

  public void addFocusListener(FocusListener listener) {
    focus.addElement(listener);
  }

  public void removeFocusListener(FocusListener listener) {
    focus.removeElement(listener);
  }

  public void addKeyListener(KeyListener listener) {
    key.addElement(listener);
  }

  public void removeKeyListener(KeyListener listener) {
    key.removeElement(listener);
  }


  // one ActionListener method

  public void actionPerformed(ActionEvent ae) {
    Object source = ae.getSource();
    //System.err.println("action event: "+ae);
    if (source == textfield) {
      setNumberFromField();
    } else {
      for (int i=0; i<count; i++) {
	if (source == dec[i]) {
	  lowerNumber(i);
	  return;
	}
	if (source == inc[i]) {
	  raiseNumber(i);
	  return;
	}
      }
    }
  }

  protected void throwActionPerformed(String command) {
    for (int i=0; i<action.size(); i++) {
      ActionEvent event;
      event = new ActionEvent(this, ActionEvent.ACTION_PERFORMED, command);
      ((ActionListener)action.elementAt(i)).actionPerformed(event);
    }
  }

  // three KeyListener methods

  public void keyPressed(KeyEvent ke) {
    // System.err.println("NumberInput.keyPressed() "+ke);
    switch (ke.getKeyCode()) {
    case KeyEvent.VK_ESCAPE:
      // cancel editing
      textfield.setText(getNumber());
      break;
    default:
      // ignore
      break;
    }
    for (int i=0; i<key.size(); i++) {
      KeyEvent event;
      event = new KeyEvent(this, ke.getID(), ke.getWhen(),
			   ke.getModifiers(), ke.getKeyCode(),
			   ke.getKeyChar());
      ((KeyListener)key.elementAt(i)).keyPressed(event);
    }
  }

  public void keyReleased(KeyEvent ke) {
    for (int i=0; i<key.size(); i++) {
      KeyEvent event;
      event = new KeyEvent(this, ke.getID(), ke.getWhen(),
			   ke.getModifiers(), ke.getKeyCode(),
			   ke.getKeyChar());
      ((KeyListener)key.elementAt(i)).keyReleased(event);
    }
  }

  public void keyTyped(KeyEvent ke) {
    for (int i=0; i<key.size(); i++) {
      KeyEvent event;
      event = new KeyEvent(this, ke.getID(), ke.getWhen(),
			   ke.getModifiers(), ke.getKeyCode(),
			   ke.getKeyChar());
      ((KeyListener)key.elementAt(i)).keyTyped(event);
    }
  }


  // two FocusListener methods plus additional Focus method

  public void focusGained(FocusEvent fe) {
    // ignore for self
    // pass on to subscribed FocusListeners
    for (int i=0; i<focus.size(); i++) {
      FocusEvent event;
      event = new FocusEvent(this, FocusEvent.FOCUS_GAINED, fe.isTemporary());
      ((FocusListener)focus.elementAt(i)).focusGained(event);
    }
  }

  public void focusLost(FocusEvent fe) {
    // set value for self
    //if (config.setOnFocusLost())
    setNumberFromField();
    // pass on to subscribed FocusListeners
    for (int i=0; i<focus.size(); i++) {
      FocusEvent event;
      event = new FocusEvent(this, FocusEvent.FOCUS_LOST, fe.isTemporary());
      ((FocusListener)focus.elementAt(i)).focusLost(event);
    }
  }

  public void requestFocus() {
    textfield.requestFocus();
  }


  // Normal interaction methods

  public void setLabel(String labeltext) {
    if (label != null) label.setText(labeltext);
  }

  public String getLabel() {
    if (label != null) return label.getText();
    else return null;
  }


  // the (il)legal part does not function well !!!
  
  /**
   * Internal reverter of setIllegal()
   */
  protected void setLegal() {
    textfield.setForeground(Color.black);
  }

  /**
   * Meant to be called by NumberInputEvent-receiver when input-value was
   * illegal. Will be unset when new input has been given (needs to be tested
   * with a test-run...).
   */
  public void setIllegal() {
    textfield.selectAll();
    textfield.setForeground(Color.red);
  }


  // overridden Container methods

  public final Component add(Component component) {
    System.err.println("An NumberInput is not a Container as such.\n"+
		       "Component not added.\n");
    return component;
  }
}
