// File: Config.java

package kero.util;

import java.util.*;
import java.io.*;
import java.awt.*; // Hm, for _printing_ colors only... override in AwtConfig?
import java.lang.reflect.*;

/**
 * Can read config files. Has an interface for several useful values from such
 * an option and its value.
 *
 * <P>One option can be specified per line. The value of the option comes after
 * an equal sign (=). Comments start with a hash (#) and end with an
 * end-of-line.
 *
 * <P>A default config file should be supplied with the package this file is
 * part of. As of 18 Dec 1999, it can be generated with
 * printConfig(). <B>However, this works only if there are functions with the
 * name ``getValue()'', where ``Value'' is the exact String as should be in the
 * config file.</b> This may not be as nice as possible, but it enforces
 * consequent naming (which is a good habit).
 *
 * @author    Kero van Gelder (setup of the config-strings)
 * @author    Rutger Nijlunsing (parsing the config-string)
 * @copyright The Authors (the GNU General Public License applies)
 * @since     21 XI 1999
 * @version   18 XII 1998
 */
public abstract class Config
{
  protected Hashtable config;

  public Config() {
    config = new Hashtable();
  }

  public void readFrom(LineReader reader) throws IOException {
    reader.skipSpace();
    while (reader.available()>0) {
      // System.err.println("available: "+reader.available()+".");
      char first = reader.peek();
      switch (first) {
      case '#':
      case '%':
	reader.skipLine();
	break;
      default:
	String key = reader.readAlphaNumericString();
	reader.passChar('=');
	reader.skipSpace();
	String value = reader.readRestOfLine();
	config.put(key,value);
	break;
      }
      reader.skipSpace();
    }
  }

  public void writeTo(String filename) throws IOException {
    FileOutputStream fos = new FileOutputStream(filename);
    PrintStream ps = new PrintStream(fos);
    Enumeration enum = config.keys();
    while (enum.hasMoreElements()) {
      String bla = (String)enum.nextElement();
      ps.println(bla+" = "+config.get(bla));
    }
  }

  public String get(String str) throws NoConfigStringException {
    String result = (String)config.get(str);
    if (result == null) {
      throw new NoConfigStringException("Config contains no "+str);
    }
    return result;
  }

  public int getInteger(String str) throws NoConfigStringException {
    String s = get(str);
    return Integer.parseInt(s);
  }

  public double getDouble(String str) throws NoConfigStringException {
    String s = get(str);
    Double result = Double.valueOf(s);
    System.err.println("Got value "+result+" for "+str);
    return Double.valueOf(s).doubleValue();
  }

  /**
   * Return true if config string str is 'yes',
   * false otherwise
   */
  public boolean getBoolean(String str) throws NoConfigStringException {
    String s = get(str);
    return "yes".equals(MyString.toLower(s).toString());
  }

  boolean isSeparator(char c) {
    return !(Character.isLowerCase(c) ||
	     Character.isUpperCase(c) ||
	     Character.isDigit(c));
  }

  /**
   * Get idx'th number from config string. Start counting at one.
   */
  public StringBuffer getStringArrayElement(String str, int idx)
    throws NoConfigStringException
  {
    StringBuffer res = new StringBuffer();
    int x = 0;			// x of current char
    int i = 1;			// Index of current char
    int len = str.length();
    // Find element idx
    while ((x < len) && (i < idx)) {
      if (isSeparator(str.charAt(x))) {
	i++;
	x++;
	// Skip separator(s)
	while ((x < len) && (isSeparator(str.charAt(x)))) {
	  x++;
	}
      } else {
	x++;
      }
    }
    if (idx < i) {
      throw new NoConfigStringException("Config item "+str+"contains no "+
					idx+"th String");
    }
    // Copy element into res
    while ((x < len) && (!isSeparator(str.charAt(x)))) {
      res.append(str.charAt(x));
      x++;
    }
    return res;
  }

  /**
   * Returns the idx-th (1..) integer element of config string str
   */
  public int getIntArrayElement(String str, int idx)
    throws NoConfigStringException
  {
    StringBuffer res = getStringArrayElement(get(str), idx);
    return Integer.parseInt(res.toString());
  }

  /**
   * Returns the idx-th (1..) boolean element of config string str
   */
  public boolean getBooleanArrayElement(String str, int idx)
    throws NoConfigStringException
  {
    String res = getStringArrayElement(get(str), idx).toString();
    return "yes".equals(MyString.toLower(res).toString());
  }

  public void printConfig(PrintStream ps) {
    Method[] method = getClass().getMethods();
    Class superclass = getClass().getSuperclass();
    Method[] supermethod = superclass.getMethods();
    for (int i=0; i<method.length; i++) {
      if ("printAllMethods".equals(method[i].getName())) continue;
      int j;
      for (j=0; j<supermethod.length; j++) {
	if (method[i].getName().equals(supermethod[j].getName())) break;
      }
      if (j==supermethod.length) try {
	String key = method[i].getName().substring(3); // based on assumption!
	Class resulttype = method[i].getReturnType();
	String resultstring;
	if (resulttype.isInstance(new Color(1,2,3))) {
	  Color c = (Color)method[i].invoke(this, null);
	  resultstring = ""+c.getRed()+", "+c.getGreen()+", "+c.getBlue();
	} else {
	  resultstring = String.valueOf(method[i].invoke(this, null));
	}
	ps.println(key+" = "+resultstring);
      } catch (Exception e) {}
    }
  }
}
