// File: Game.java

package jvpc.vga;

import java.util.*;
import java.io.*;
import kero.util.*;

/**
 * <H1>The JVPC Internals</h1>
 *
 * VGA-Planets is a PBEM made by Tim Wisseman. It knows up to 11 players, each
 * playing another race. Diplomacy, economics and warfare are the major parts
 * of this game.
 *
 * <H2>This is the Game.java</h2>
 *
 * <P>Though the concept and balance of the game is great, the software by Tim
 * sucks (excuse me for saying so). Whereas his dataformat was not as well
 * protected as he might have thought and the client was (and still is) cheap,
 * VGA-Planets exhibits a lot of <A HREF="http://www.opensource.org">Open
 * Source</a> features.
 *
 * <P>So lots of utilities and even full featured clients exist. Yet no client
 * can handle team-games correctly. The emphasis for this Java client for me
 * is on properly reading, editing and writing data for multiple races in one
 * game. At the same time, some kind of database must be designed to remember
 * all data there is to remember (without arbitrary restictions, like ten-turns
 * history only)
 *
 * <P>For other information, see ../doc/index.html
 * 
 * @author    Kero van Gelder
 * @copyright Kero van Gelder (the GNU General Public License applies)
 * @since     10 VII 1998
 * @version   14 III 2001
 */
public class Game
  extends Reporter
  implements VGAPlanets
{
  StringBuffer workdir,globaldir;
  public InputFileOpener ifo;
  public OutputFileOpener ofo;

  int rstCount, trnCount;
  int currentTurn; 
  Vector rst; // rst[1..currentTurn] are used; rst[0] as well
  Vector trn; // trn[1..currentTurn] are used
  Hull[] hull;
  Truehull[] truehull;
  Hullfunc hullfunc;
  Engine[] engine;
  Torpedo[] torpedo;
  Beam[] beam;
  Vector missvec;
  Hashtable misshash;
  String[] racelong,raceshort,raceadj;
  byte[] fizz; // registry key

  boolean locked = false;

  Reportable reportable;

  /**
   * Constructor with no work-directory given. Assume player wants to play in
   * the current directory.
   */
  public Game() {
    this(".",".");
  }

  /**
   * Constructor with the work-directory given. Assume that the global
   * directory is the current directory.
   */
  public Game(String workdir) {
    this(workdir,".");
  }

  private void setFileOpeners() {
    if (workdir.charAt(workdir.length()-1)!='/') workdir.append("/");
    if (globaldir.charAt(globaldir.length()-1)!='/') globaldir.append("/");
    try {
      ifo=new InputFileOpener(workdir.toString(),globaldir.toString());
    } catch (IOException ioe) {
      crash(11, "Directory does not exist: "+ioe);
    }
    ofo=new OutputFileOpener(workdir.toString(),globaldir.toString());
  }

  /**
   * Constructor with both directories.
   */
  public Game(String workdir,String globaldir) {
    if (workdir==null) crash(1000, "No work directory given");
    if (globaldir==null) crash(1000, "No global directory given");
    this.workdir=new StringBuffer(workdir);
    this.globaldir=new StringBuffer(globaldir);
    setFileOpeners();

    rst=new Vector();
    trn=new Vector();
    hull=new Hull[hulls+1];
    truehull=new Truehull[races+1];
    hullfunc=new Hullfunc();
    engine=new Engine[engines+1];
    torpedo=new Torpedo[torpedoes+1];
    beam=new Beam[beams+1];
    misshash=new Hashtable();
    racelong=new String[races+1];
    raceshort=new String[races+1];
    raceadj=new String[races+1];
  }


  /**
   * Read global data (that is, data that does not change in the course of
   * time. Concerning starmaps and spec-files, changes occur only once: from
   * zero to a meaningful number.)
   */
  private void readGlobal() {
    InputStream infile;

    // read jvpc.cfg
    try {
      infile=ifo.open("jvpc.cfg",false);
      LineReader source=new LineReader(infile);
    } catch (IOException ioe) {
      // no jvpc config file found, ignore
    }

    // read xyplan(??).dat and planet.nm into rst[0]
    rst.addElement(new Universe(this,0,RstOrTrn.RST)); // the first, so it's placed at index 0

    // read explorer files xyplan??.dat (grows, to Universe?)
    int mapsFound=0;
    for (int i=1;i<=races;i++) {
      try {
	infile=ifo.open("xyplan"+i+".dat",false); // only this line throws
	XYPlanGen xyplangen=new XYPlanGen(infile);
	xyplangen.generateAll((Universe)rst.elementAt(0));
	mapsFound++;
      } catch (IOException ioe) {} // ignore, file does not exist
    }

    // when no explorer files were found, read xyplan.dat
    if (mapsFound==0) {
      try {
	infile=ifo.open("xyplan.dat",true); // only this line throws
	XYPlanGen xyplangen=new XYPlanGen(infile);
	xyplangen.generateAll((Universe)rst.elementAt(0));
      } catch (IOException ioe) {
	crash(11, "No starmap found: "+ioe+"\nThe explorer-generated files "+
	      "xyplan??.dat were not found, either");
      }
    }

    // read planet.nm, after xyplan(??).dat has been read
    try {
      infile=ifo.open("planet.nm",true);
      PlanetNMGen planetnmgen=new PlanetNMGen(infile);
      planetnmgen.generateAll((Universe)rst.elementAt(0));
    } catch(IOException ioe) {
      crash(11, "File planet.nm not found: "+ioe);
    }

    // read hullspec.dat
    try {
      infile=ifo.open("hullspec.dat",true); // Universe? ...
      HullGen hullgen=new HullGen(infile);
      hullgen.generateAll(this);
    } catch (IOException ioe) {
      crash(11, "File hullspec.dat not found: "+ioe);
    }

    // read hullfunc.dat
    try {
      infile=ifo.open("hullfunc.dat",true); // Universe? ...
      HullfuncDatGen hullfuncdatgen=new HullfuncDatGen(infile);
      hullfuncdatgen.generateAll(this);
    } catch (IOException ioe) {
      // no problem if file does not exist (defaults will be used)
    }

    // read truehull.dat
    try {
      infile=ifo.open("truehull.dat",true); // Universe? ...
      TruehullGen truehullgen=new TruehullGen(infile);
      truehullgen.generateAll(this);
    } catch (IOException ioe) {
      crash(11, "File truehull.dat not found: "+ioe);
    }

    // read engspec.dat
    try {
      infile=ifo.open("engspec.dat",true); // Universe? ...
      EngineGen enginegen=new EngineGen(infile);
      enginegen.generateAll(this);
    } catch (IOException ioe) {
      crash(11, "File engspec.dat not found: "+ioe);
    }

    // read torpspec.dat
    try {
      infile=ifo.open("torpspec.dat",true); // Universe? ...
      TorpedoGen torpedogen=new TorpedoGen(infile);
      torpedogen.generateAll(this);
    } catch (IOException ioe) {
      crash(11, "File torpspec.dat not found: "+ioe);
    }

    // read beamspec.dat
    try {
      infile=ifo.open("beamspec.dat",true); // Universe? ...
      BeamGen beamgen=new BeamGen(infile);
      beamgen.generateAll(this);
    } catch (IOException ioe) {
      crash(11, "File beamspec.dat not found: "+ioe);
    }

    // read race.nm (can change, to Universe?)
    try {
      infile=ifo.open("race.nm",true); // Universe? Player! ...
      RaceNMGen racenmgen=new RaceNMGen(infile);
      racenmgen.generateAll(this);
    } catch (IOException ioe) {
      crash(11, "File race.nm not found: "+ioe);
    }

    // setup normal missions
    missvec=Mission.setupTimMissions();
    for (int i=0; i<missvec.size(); i++) {
      Mission m=(Mission)missvec.elementAt(i);
      misshash.put(m.getName(),m);
    }
    // read mission.ini (can change?, to Universe?)
    try {
      infile=ifo.open("mission.ini",true); // Universe? Player! ...
      LineReader source=new LineReader(infile);
      MissionGen missiongen=new MissionGen();
      missiongen.generateExtendedMissions(this,source); // calls putMission()
    } catch (IOException ioe) {
      // mission.ini not avaiable, probably TimHost game
    }

    // Read fizz.bin (registration key); TRN will override, if we are shareware
    try {
      infile=ifo.open("fizz.bin",true);
      fizz = new FizzBinGen(infile).generate();
    } catch(IOException ioe) {
      report("Missing file ``FIZZ.BIN'', assuming shareware version: "+ioe);
      fizz = FizzBinGen.sharewareRegKey;
    }
  }

  public void putMission(Mission m) {
    int index=m.getId();
    if ((missvec.size()>index) && (missvec.elementAt(index)!=null))
      warn("Already mission with that index: "+index);
    String key=m.getName();
    if (misshash.get(key)!=null) warn("Already mission with that name: "+key);
    if (index>=missvec.size()) missvec.setSize(index+1);
    missvec.setElementAt(m,index);
    misshash.put(key,m);
  }

    /**
     * Get the size of the vector containing the missions; upper limit on
     * amount and gives a handle to retrieve all Missions.
     */
    public int getMissVecSize() {
	return missvec.size();
    }

  public Mission getMission(int i) {
    if (i==UNKNOWN) return null;
    if ((i<0) || (i>=missvec.size())) return null;
    return (Mission)missvec.elementAt(i);
  }

  public Mission getMission(String str) {
    if ((str==null) || str.equals("")) return null;
    return (Mission)misshash.get(str);
  }

  public String getLongRace(int player) {
    if ((player<0) || (player>races)) return null;
    if (player==0) return "None";
    return racelong[player];
  }

  /**
   * Read all data from the working directory, i.e.\ where the non-global
   * (player) files are. As programmer, you probably want to call readAll().
   */
  public void readWork(boolean[] wantsToPlay) {
    InputStream infile;
    Gen[] gen=new Gen[races+1];
    boolean anyplayers=false;

    // do not read own history; done in another way

    // read results (can be a rehost; compare with history)
    for (int player=1;player<=races;player++) {
      if (wantsToPlay[player]) try {
	InputStream fis=ifo.open("player"+player+".rst",true);
	RstGen rstgen=new RstGen(fis);
	int turnnr=rstgen.getTurnNumber();
	Universe u=getNoHistoryUniverse(turnnr,RstOrTrn.RST); // from rst[]
	if (u==null) {
	  u=new Universe((Universe)rst.elementAt(0),turnnr,RstOrTrn.RST);
	  //u.setTurnNumber(turnnr);
	  //u.set(RstOrTrn.RST);
	  putUniverse(u,turnnr);
	  u.readConfig(ifo); // only PHost pconfig.src, not TimHost config
	}
	anyplayers=true;
	// scan buffer second time for complete info
	fis=ifo.open("player"+player+".rst",true);
	rstgen=new RstGen(fis); // need to reset the stream
	u.setReportable(this); // cosmetic
	u.readRst(rstgen,player);
      } catch (IOException ioe) {
	// no result, ignore
      }
    }

    // check gen-files
    // if available, read it and all accompanying files
    for (int player=1; player<=races; player++) {
      if (wantsToPlay[player]) try {
	infile=ifo.open("gen"+player+".dat",false);
	try {
	  GenGen gengen=new GenGen(infile,player,true);
	  gen[player]=gengen.generate();
	  int turnnr=gen[player].getTurnNumber();
	  Universe u=getNoHistoryUniverse(turnnr,RstOrTrn.RST); // from rst[]
	  if (u==null) {
	    u=new Universe((Universe)rst.elementAt(0),turnnr,RstOrTrn.RST);
	    //u.setTurnNumber(turnnr);
	    //u.set(RstOrTrn.RST);
	    putUniverse(u,turnnr);
	  }
	  if (u.getGen(player)==null) {
	    u.putGen(gen[player],player);
	    u.setReportable(this); // cosmetic
	    u.readDisData(ifo,player);
	  }
	  anyplayers=true;
	} catch(IOException ioe2) {
	  ioe2.printStackTrace();
	  crash(11, "\nFatal error while reading gen"+player+".dat");
	}
      } catch(IOException ioe) {
	// ignore, gen??.dat not found
      }
    }

    // any players?
    if (!anyplayers) {
      System.err.println("No playerdata found");
      return;
    }

    // read util??.dat after rst's and normal data
    // it can belong to either, both or none :-|
    for (int player=1;player<=races;player++) {
      if (wantsToPlay[player]) try {
	infile=ifo.open("util"+player+".dat",false);
	UtilGen utilgen=new UtilGen(this,infile);
	int turnnr=utilgen.getTurnNumber();
	// System.err.println("UTIL from turn "+turnnr);
	// records of UTIL in proper order, no need to reset ifo
	Universe u=getNoHistoryUniverse(turnnr,RstOrTrn.RST);
	if (u!=null) utilgen.generateAll(u,player);
      } catch(IOException ioe) {
	// no UTIL (no PHost?), ignore
      }
    }

      // read dat-files (can contain new info, set by other utils)
      for (int player=1; player<=races; player++) {
        if (wantsToPlay[player]) if (gen[player]!=null) {
	    //System.err.println("gen["+player+"] found");
	    int turnnr=gen[player].getTurnNumber();
	    Universe u=getNoHistoryUniverse(turnnr,RstOrTrn.TRN); // from trn[]
	    if (u==null) {
		report("Copying rst-universe to trn-universe");
		Universe urst=getNoHistoryUniverse(turnnr,RstOrTrn.RST);
		if (urst!=null) {
		    u=urst.toTrnUniverse();
		} else {
		    report("Creating new TRN universe");
		    u=new Universe((Universe)rst.elementAt(0),turnnr,RstOrTrn.TRN);
		    //u.setTurnNumber(turnnr);
		    //u.set(RstOrTrn.TRN);
		}
		putUniverse(u,turnnr);
	    }
	    u.setReportable(this); // cosmetic
	    u.readDatData(ifo,player);
	    wantsToPlay[player] = false; // got all his data...
        }
      }

      // if no dat-files were found, read player??.trn that fit a result universe
      for (int player=1;player<=races;player++) {
	if (/*still*/wantsToPlay[player]) try {
	    infile=ifo.open("player"+player+".trn",false);
	    TrnGen trngen=new TrnGen(infile);
	    String ts=trngen.getTimestamp(player);
	    System.err.println("TRN from turn "+ts);
	    Universe u=getNoHistoryUniverse(ts,RstOrTrn.TRN);
	    if (u==null) {
		Universe urst=getNoHistoryUniverse(ts,RstOrTrn.RST);
		if (urst!=null) {
		    u=urst.toTrnUniverse();
		    putUniverse(u,u.getTurnNumber());
		}
	    }
  	    trngen.generateAll(u,player);
        } catch(IOException ioe) {
	    // no TRN file, ignore
        }
      }

    // if still no current trn-universe (fresh RST?) copy it from rst-universe
    if (getNoHistoryUniverse(currentTurn,RstOrTrn.TRN)==null) {
      Universe rstUniv=getNoHistoryUniverse(currentTurn,RstOrTrn.RST);
      Universe trnUniv=rstUniv.toTrnUniverse();
      putUniverse(trnUniv,currentTurn);
    }
  }

  /**
   * Read all data there is for this game. That includes history, current turn
   * (and future, once we know how to handle it; prepare host-data, run a host,
   * etc). In any case, find out what can be read, what must be read, without
   * ``forgetting'' any (externally modified?) data, without ambiguities. If
   * necessary for speed, read only part of the data; read other things later.
   */
  public void readAll(boolean[] wantsToPlay) {
    readGlobal();
    readWork(wantsToPlay);
    report("Read "+Planet.realPlanets()+" (explored) planets");
    report("All data loaded");
    checkConsistency();
    // warn for unsupported shareware
    if (fizz.equals(FizzBinGen.sharewareRegKey)) {
      report("JVPC does not restrict itself when playing shareware");
      report("   usage of another client recommended");
      report("   or copy FIZZ.BIN, so JVPC knows the registration");
      warn("JVPC does not restrict itself when playing shareware");
      warn("   usage of another client recommended.");
      warn("   or copy FIZZ.BIN, so JVPC knows the registration");
    }
  }

  private void updateFromHistory(int turnnr) {
    // System.err.println("updateFromHistory("+turnnr+")");
    String dir=workdir.toString()+turnnr+File.separator;
    ifo.setWorkDir(dir);
    try {
      ifo=new InputFileOpener(dir,dir);
      boolean[] wantsToPlay = new boolean[races+1];
      for (int i=1; i<=races; i++) wantsToPlay[i] = true;
      readWork(wantsToPlay); // read from history
    } catch (IOException ioe) {
      // could not open directory
    }
    // restore old path in ifo
    dir=workdir.toString();
    ifo.setWorkDir(dir);
  }

  public void writeAll() {
      unsetReportable(); // <<<--- !!!
      checkConsistency();
      ((Universe)rst.elementAt(currentTurn)).writeAll(ifo,ofo);
      ((Universe)trn.elementAt(currentTurn)).writeAll(ifo,ofo);
      makeTurn();
      report("All data saved.");
  }

  public void makeTurn() {
    Universe rstUniv = getUniverse(getCurrentTurn(), RstOrTrn.RST);
    Universe trnUniv = getUniverse(getCurrentTurn(), RstOrTrn.TRN);
    int ptr[][] = new int[races+1][500*100];
    LSBWriter comm[] = new LSBWriter[races+1];
    int commCount[] = new int[races+1];
    LSBWriter trnFile[] = new LSBWriter[races+1];
    int csAddition[] = new int[races+1];
    int cs[] = new int[races+1];
    try {
      for (int j=1; j<=races; j++) {
	if (trnUniv.getGen(j) != null) {  // isPlaying()...
	  comm[j] = new LSBWriter(new ByteArrayOutputStream(10*1024));
	}
      }

      /* commands for all ships */
      System.out.println("Checking ships...");
      for (int i=1;i<=ships;i++) {
	  Ship rstS = rstUniv.getShip(i);
	  if ((rstS != null) && (trnUniv.getGen(rstS.owner) != null)) {
	      //System.out.println("Checking ship "+i+": "+rstS.getReport());
	      Ship trnS = trnUniv.getShip(i);
	      int owner = trnS.owner;
	      if (!(rstS.getFC().equals(trnS.getFC()))) {
		  //NewCommand(7,1);
		  ptr[owner][commCount[owner]] = comm[owner].getPosition();
		  commCount[owner] += 1;
		  comm[owner].writeShort(1); // ship FC command ID
		  comm[owner].writeShort(trnS.id);
		  comm[owner].writeString(trnS.fc);
	      }
	      if (rstS.getWarp() != trnS.getWarp()) {
		// NewCommand(6,2);
		ptr[owner][commCount[owner]] = comm[owner].getPosition();
		commCount[owner] += 1;
		comm[owner].writeShort(2); // ship warp command ID
		comm[owner].writeShort(trnS.id);
		comm[owner].writeShort(trnS.getWarp());
	      }
	      if ((rstS.getDX()!=trnS.getDX()) || (rstS.getDY()!=trnS.getDY())) {
		// NewCommand(8,3);
		ptr[owner][commCount[owner]] = comm[owner].getPosition();
		commCount[owner] += 1;
		comm[owner].writeShort(3);
		comm[owner].writeShort(trnS.id);
		comm[owner].writeShort(trnS.dx);
		comm[owner].writeShort(trnS.dy);
	      }
	      if (rstS.mission != trnS.mission) {
		// NewCommand(6,4);
		ptr[owner][commCount[owner]] = comm[owner].getPosition();
		commCount[owner] += 1;
		comm[owner].writeShort(4);
		comm[owner].writeShort(trnS.id);
		comm[owner].writeShort(trnS.mission);
	      }
	      if (rstS.enemy != trnS.enemy) {
		// NewCommand(6,5);
		ptr[owner][commCount[owner]] = comm[owner].getPosition();
		commCount[owner] += 1;
		comm[owner].writeShort(5);
		comm[owner].writeShort(trnS.id);
		comm[owner].writeShort(trnS.enemy);
	      }
	      if (rstS.towtarget != trnS.towtarget) {
		// NewCommand(6,6);
		ptr[owner][commCount[owner]] = comm[owner].getPosition();
		commCount[owner] += 1;
		comm[owner].writeShort(6);
		comm[owner].writeShort(trnS.id);
		comm[owner].writeShort(trnS.towtarget);
	      }
	      if (rstS.inttarget != trnS.inttarget) {
		// NewCommand(6,10);
		ptr[owner][commCount[owner]] = comm[owner].getPosition();
		commCount[owner] += 1;
		comm[owner].writeShort(10);
		comm[owner].writeShort(trnS.id);
		comm[owner].writeShort(trnS.inttarget);
	      }
	      if (!(rstS.getName().equals(trnS.getName()))) {
		// NewCommand(24,7);
		ptr[owner][commCount[owner]] = comm[owner].getPosition();
		commCount[owner] += 1;
		comm[owner].writeShort(7);
		comm[owner].writeShort(trnS.id);
		comm[owner].writeString(trnS.name);
	      }
	      if (!(rstS.toplanet.equals(trnS.toplanet))) {
		// NewCommand(18,8);
		ptr[owner][commCount[owner]] = comm[owner].getPosition();
		commCount[owner] += 1;
		comm[owner].writeShort(8);
		comm[owner].writeShort(trnS.id);
		comm[owner].writeShort(trnS.toplanet.cargo[Cargo.NEUTRONIUM].amount);
		comm[owner].writeShort(trnS.toplanet.cargo[Cargo.TRITANIUM].amount);
		comm[owner].writeShort(trnS.toplanet.cargo[Cargo.DURANIUM].amount);
		comm[owner].writeShort(trnS.toplanet.cargo[Cargo.MOLYBDENIUM].amount);
		comm[owner].writeShort(trnS.toplanet.cargo[Cargo.COLCLANS].amount);
		comm[owner].writeShort(trnS.toplanet.cargo[Cargo.SUPPLIES].amount);
		comm[owner].writeShort(trnS.getLocation().getPlanet().getId());
	      }
	      if ((rstS.toship.shipid != trnS.toship.shipid)||
		  (rstS.toship.getNeutronium()!=trnS.toship.getNeutronium())||
		  (rstS.toship.getTritanium()!=trnS.toship.getTritanium())||
		  (rstS.toship.getDuranium()!=trnS.toship.getDuranium())||
		  (rstS.toship.getMolybdenium()!=trnS.toship.getMolybdenium())||
		  (rstS.toship.getColClans()!=trnS.toship.getColClans())||
		  (rstS.toship.getSupplies()!=trnS.toship.getSupplies())) {
		// NewCommand(18,9);
		ptr[owner][commCount[owner]] = comm[owner].getPosition();
		commCount[owner] += 1;
		comm[owner].writeShort(9);
		comm[owner].writeShort(trnS.id);
		comm[owner].writeShort(trnS.toship.getNeutronium());
		comm[owner].writeShort(trnS.toship.getTritanium());
		comm[owner].writeShort(trnS.toship.getDuranium());
		comm[owner].writeShort(trnS.toship.getMolybdenium());
		comm[owner].writeShort(trnS.toship.getColClans());
		comm[owner].writeShort(trnS.toship.getSupplies());
		comm[owner].writeShort(trnS.toship.shipid);
	      }
	      if (rstS.getNeutronium() != trnS.getNeutronium()) {
		// NewCommand(6,11);
		ptr[owner][commCount[owner]] = comm[owner].getPosition();
		commCount[owner] += 1;
		comm[owner].writeShort(11);
		comm[owner].writeShort(trnS.id);
		comm[owner].writeShort(trnS.getNeutronium());
	      }
	      if (rstS.getTritanium() != trnS.getTritanium()) {
		// NewCommand(6,12);
		ptr[owner][commCount[owner]] = comm[owner].getPosition();
		commCount[owner] += 1;
		comm[owner].writeShort(12);
		comm[owner].writeShort(trnS.id);
		comm[owner].writeShort(trnS.getTritanium());
	      }
	      if (rstS.getDuranium() != trnS.getDuranium()) {
		// NewCommand(6,13);
		ptr[owner][commCount[owner]] = comm[owner].getPosition();
		commCount[owner] += 1;
		comm[owner].writeShort(13);
		comm[owner].writeShort(trnS.id);
		comm[owner].writeShort(trnS.getDuranium());
	      }
	      if (rstS.getMolybdenium() != trnS.getMolybdenium()) {
		// NewCommand(6,14);
		ptr[owner][commCount[owner]] = comm[owner].getPosition();
		commCount[owner] += 1;
		comm[owner].writeShort(14);
		comm[owner].writeShort(trnS.id);
		comm[owner].writeShort(trnS.getMolybdenium());
	      }
	      if (rstS.getSupplies() != trnS.getSupplies()) {
		// NewCommand(6,15);
		ptr[owner][commCount[owner]] = comm[owner].getPosition();
		commCount[owner] += 1;
		comm[owner].writeShort(15);
		comm[owner].writeShort(trnS.id);
		comm[owner].writeShort(trnS.getSupplies());
	      }
	      if (rstS.getColClans() != trnS.getColClans()) {
		// NewCommand(6,16);
		ptr[owner][commCount[owner]] = comm[owner].getPosition();
		commCount[owner] += 1;
		comm[owner].writeShort(16);
		comm[owner].writeShort(trnS.id);
		comm[owner].writeShort(trnS.getColClans());
	      }
	      if (rstS.getAmmo() != trnS.getAmmo()) {
		// NewCommand(6,17);
		ptr[owner][commCount[owner]] = comm[owner].getPosition();
		commCount[owner] += 1;
		comm[owner].writeShort(17);
		comm[owner].writeShort(trnS.id);
		comm[owner].writeShort(trnS.getAmmo());
	      }
	      if (rstS.getMC() != trnS.getMC()) {
		// NewCommand(6,18);
		ptr[owner][commCount[owner]] = comm[owner].getPosition();
		commCount[owner] += 1;
		comm[owner].writeShort(18);
		comm[owner].writeShort(trnS.id);
		comm[owner].writeShort(trnS.getMC());
	      }
	  }
      }

      /* commands for all planets */
      System.out.println("Checking planets...");
      for (int i=1; i<=planets; i++) {
	  Planet rstP = rstUniv.getPlanet(i);
	  if ((rstP != null) && (trnUniv.getGen(rstP.owner) != null)) {
	      //System.out.println("Checking planet "+i+": "+rstP.getReport());
	      Planet trnP = trnUniv.getPlanet(i);
	      int owner = trnP.owner;
	      if (!(rstP.getFC().equals(trnP.getFC()))) {
		  // NewCommand(7,21);
		  ptr[owner][commCount[owner]] = comm[owner].getPosition();
		  commCount[owner] += 1;
		  comm[owner].writeShort(21);
		  comm[owner].writeShort(trnP.id);
		  comm[owner].writeString(trnP.getFC());
	      }
	      if (rstP.getMines() != trnP.getMines()) {
		  // NewCommand(6,22);
		  ptr[owner][commCount[owner]] = comm[owner].getPosition();
		  commCount[owner] += 1;
		  comm[owner].writeShort(22);
		  comm[owner].writeShort(trnP.id);
		  comm[owner].writeShort(trnP.getMines());
	      }
	      if (rstP.getFactories() != trnP.getFactories()) {
		  // NewCommand(6,23);
		  ptr[owner][commCount[owner]] = comm[owner].getPosition();
		  commCount[owner] += 1;
		  comm[owner].writeShort(23);
		  comm[owner].writeShort(trnP.id);
		  comm[owner].writeShort(trnP.getFactories());
	      }
	      if (rstP.getDefense() != trnP.getDefense()) {
		  // NewCommand(6,24);
		  ptr[owner][commCount[owner]] = comm[owner].getPosition();
		  commCount[owner] += 1;
		  comm[owner].writeShort(24);
		  comm[owner].writeShort(trnP.id);
		  comm[owner].writeShort(trnP.getDefense());
	      }
	      if (rstP.getNeutronium() != trnP.getNeutronium()) {
		  // NewCommand(8,25);
		  ptr[owner][commCount[owner]] = comm[owner].getPosition();
		  commCount[owner] += 1;
		  comm[owner].writeShort(25);
		  comm[owner].writeShort(trnP.id);
		  comm[owner].writeInt(trnP.getNeutronium());
	      }
	      if (rstP.getTritanium() != trnP.getTritanium()) {
		  // NewCommand(8,26);
		  ptr[owner][commCount[owner]] = comm[owner].getPosition();
		  commCount[owner] += 1;
		  comm[owner].writeShort(26);
		  comm[owner].writeShort(trnP.id);
		  comm[owner].writeInt(trnP.getTritanium());
	      }
	      if (rstP.getDuranium() != trnP.getDuranium()) {
		  // NewCommand(8,27);
		  ptr[owner][commCount[owner]] = comm[owner].getPosition();
		  commCount[owner] += 1;
		  comm[owner].writeShort(27);
		  comm[owner].writeShort(trnP.id);
		  comm[owner].writeInt(trnP.getDuranium());
	      }
	      if (rstP.getMolybdenium() != trnP.getMolybdenium()) {
		  // NewCommand(8,28);
		  ptr[owner][commCount[owner]] = comm[owner].getPosition();
		  commCount[owner] += 1;
		  comm[owner].writeShort(28);
		  comm[owner].writeShort(trnP.id);
		  comm[owner].writeInt(trnP.getMolybdenium());
	      }
	      if (rstP.getColClans() != trnP.getColClans()) {
		  // NewCommand(8,29);
		  ptr[owner][commCount[owner]] = comm[owner].getPosition();
		  commCount[owner] += 1;
		  comm[owner].writeShort(29);
		  comm[owner].writeShort(trnP.id);
		  comm[owner].writeInt(trnP.getColClans());
	      }
	      if (rstP.getSupplies() != trnP.getSupplies()) {
		  // NewCommand(8,30);
		  ptr[owner][commCount[owner]] = comm[owner].getPosition();
		  commCount[owner] += 1;
		  comm[owner].writeShort(30);
		  comm[owner].writeShort(trnP.id);
		  comm[owner].writeInt(trnP.getSupplies());
	      }
	      if (rstP.getMC() != trnP.getMC()) {
		  // NewCommand(8,31);
		  ptr[owner][commCount[owner]] = comm[owner].getPosition();
		  commCount[owner] += 1;
		  comm[owner].writeShort(31);
		  comm[owner].writeShort(trnP.id);
		  comm[owner].writeInt(trnP.getMC());
	      }
	      if (rstP.getColTax() != trnP.getColTax()) {
		  // NewCommand(6,32);
		  ptr[owner][commCount[owner]] = comm[owner].getPosition();
		  commCount[owner] += 1;
		  comm[owner].writeShort(32);
		  comm[owner].writeShort(trnP.id);
		  comm[owner].writeShort(trnP.getColTax());
	      }
	      if (rstP.getNatTax() != trnP.getNatTax()) {
		  // NewCommand(6,33);
		  ptr[owner][commCount[owner]] = comm[owner].getPosition();
		  commCount[owner] += 1;
		  comm[owner].writeShort(33);
		  comm[owner].writeShort(trnP.id);
		  comm[owner].writeShort(trnP.getNatTax());
	      }
	      if (trnP.buildingBase()) {
		  // NewCommand(4,34);
		  ptr[owner][commCount[owner]] = comm[owner].getPosition();
		  commCount[owner] += 1;
		  comm[owner].writeShort(34);
		  comm[owner].writeShort(trnP.id);
	      }
	  }
      }

      /* commands for bases */
      System.out.println("Checking bases...");
      for (int i=1; i<=planets; i++) {
	  BData rstB = rstUniv.getBData(i);
	  if ((rstB != null) && (trnUniv.getGen(rstB.owner) != null)) {
	      //System.out.println("Checking base "+i+": "+rstB.getReport());
	      BData trnB = trnUniv.getBData(i);
	      int owner = trnB.owner;

	      /*	      ba=rstB;
			      printf("old %3d: %3d %3d [%2d,%2d,%2d,%2d]",
			      i, ba.nofighters,ba.defense,ba.engtech,ba.hulltech,ba.beamtech,ba.torptech);
			      printf("  %d",ba.bhulltype);
			      printf("  %d",ba.fixrec);
			      printf("  %d",ba.target);
			      printf("  %d",ba.mission);
			      printf("\n");
			      
			      ba=trnB;
			      printf("new %3d: %3d %3d [%2d,%2d,%2d,%2d]",
			      i, ba.nofighters,ba.defense,ba.engtech,ba.hulltech,ba.beamtech,ba.torptech);
			      printf("  %d",ba.bhulltype);
			      printf("  %d",ba.fixrec);
			      printf("  %d",ba.target);
			      printf("  %d",ba.mission);
			      printf("\n");
			      
			      #ifdef test
			      fprintf(stderr,"%ld",i%10);
			      #endif */

	      if (rstB.getBaseDefense() != trnB.getBaseDefense()) {
		  // NewCommand(6,40);
		  ptr[owner][commCount[owner]] = comm[owner].getPosition();
		  commCount[owner] += 1;
		  comm[owner].writeShort(40);
		  comm[owner].writeShort(trnB.id);
		  comm[owner].writeShort(trnB.getBaseDefense());
	      }
	      if (rstB.getEngineTech() != trnB.getEngineTech()) {
		  // NewCommand(6,41);
		  ptr[owner][commCount[owner]] = comm[owner].getPosition();
		  commCount[owner] += 1;
		  comm[owner].writeShort(41);
		  comm[owner].writeShort(trnB.id);
		  comm[owner].writeShort(trnB.getEngineTech());
	      }
	      if (rstB.getHullTech() != trnB.getHullTech()) {
		  // NewCommand(6,42);
		  ptr[owner][commCount[owner]] = comm[owner].getPosition();
		  commCount[owner] += 1;
		  comm[owner].writeShort(42);
		  comm[owner].writeShort(trnB.id);
		  comm[owner].writeShort(trnB.getHullTech());
	      }
	      if (rstB.getBeamTech() != trnB.getBeamTech()) {
		  // NewCommand(6,43);
		  ptr[owner][commCount[owner]] = comm[owner].getPosition();
		  commCount[owner] += 1;
		  comm[owner].writeShort(43);
		  comm[owner].writeShort(trnB.id);
		  comm[owner].writeShort(trnB.getBeamTech());
	      }
	      int build = 0;
	      for (int j=1;j<=engines;j++)
		  if (rstB.getEngines(j)!=trnB.getEngines(j)) build++;
	      if (build > 0) {
		  // NewCommand(22,44);
		  ptr[owner][commCount[owner]] = comm[owner].getPosition();
		  commCount[owner] += 1;
		  comm[owner].writeShort(44);
		  comm[owner].writeShort(trnB.id);
		  for (int j=1;j<=engines;j++) { comm[owner].writeShort(trnB.getEngines(j)); }
	      }
	      build = 0;
	      for (int j=1;j<=buildhulls;j++)
		  if (rstB.getHulls(j)!=trnB.getHulls(j)) build++;
	      if (build > 0) {
		  // NewCommand(44,45);
		  ptr[owner][commCount[owner]] = comm[owner].getPosition();
		  commCount[owner] += 1;
		  comm[owner].writeShort(45);
		  comm[owner].writeShort(trnB.id);
		  for (int j=1;j<=buildhulls;j++) { comm[owner].writeShort(trnB.getHulls(j)); }
	      }
	      build = 0;
	      for (int j=1;j<=beams;j++)
		  if (rstB.getBeams(j)!=trnB.getBeams(j)) build++;
	      if (build > 0) {
		  // NewCommand(24,46);
		  ptr[owner][commCount[owner]] = comm[owner].getPosition();
		  commCount[owner] += 1;
		  comm[owner].writeShort(46);
		  comm[owner].writeShort(trnB.id);
		  for (int j=1;j<=beams;j++) { comm[owner].writeShort(trnB.getBeams(j)); }
	      }
	      build = 0;
	      for (int j=1;j<=torpedoes;j++)
		  if (rstB.getLaunchers(j)!=trnB.getLaunchers(j)) build++;
	      if (build > 0) {
		  // NewCommand(24,47);
		  ptr[owner][commCount[owner]] = comm[owner].getPosition();
		  commCount[owner] += 1;
		  comm[owner].writeShort(47);
		  comm[owner].writeShort(trnB.id);
		  for (int j=1;j<=torpedoes;j++) { comm[owner].writeShort(trnB.getLaunchers(j)); }
	      }
	      build = 0;
	      for (int j=1;j<=torpedoes;j++)
		  if (rstB.getTorpedoes(j)!=trnB.getTorpedoes(j)) build++;
	      if (build > 0) {
		  // NewCommand(24,48);
		  ptr[owner][commCount[owner]] = comm[owner].getPosition();
		  commCount[owner] += 1;
		  comm[owner].writeShort(48);
		  comm[owner].writeShort(trnB.id);
		  for (int j=1;j<=torpedoes;j++) { comm[owner].writeShort(trnB.getTorpedoes(j)); }
	      }
	      if (rstB.getFighters() != trnB.getFighters()) {
		  // NewCommand(6,49);
		  ptr[owner][commCount[owner]] = comm[owner].getPosition();
		  commCount[owner] += 1;
		  comm[owner].writeShort(49);
		  comm[owner].writeShort(trnB.id);
		  comm[owner].writeShort(trnB.getFighters());
	      }
	      if (rstB.getTarget() != trnB.getTarget()) {
		  // NewCommand(6,50);
		  ptr[owner][commCount[owner]] = comm[owner].getPosition();
		  commCount[owner] += 1;
		  comm[owner].writeShort(50);
		  comm[owner].writeShort(trnB.id);
		  comm[owner].writeShort(trnB.target);
	      }
	      if (rstB.getFixRec() != trnB.getFixRec()) {
		  // NewCommand(6,51);
		  ptr[owner][commCount[owner]] = comm[owner].getPosition();
		  commCount[owner] += 1;
		  comm[owner].writeShort(51);
		  comm[owner].writeShort(trnB.id);
		  comm[owner].writeShort(trnB.getFixRec());
	      }
	      if (rstB.getMission() != trnB.getMission()) {
		  // NewCommand(6,52);
		  ptr[owner][commCount[owner]] = comm[owner].getPosition();
		  commCount[owner] += 1;
		  comm[owner].writeShort(52);
		  comm[owner].writeShort(trnB.id);
		  comm[owner].writeShort(trnB.mission);
	      }
	      if ( ((trnB.bhulltype!=0) && (trnB.bengtype!=0)) /* &&
								  ((rstB.bhulltype!=trnB.bhulltype) || (rstB.bengtype!=trnB.bengtype))*/ ) {
		  // NewCommand(18,53);
		  ptr[owner][commCount[owner]] = comm[owner].getPosition();
		  commCount[owner] += 1;
		  comm[owner].writeShort(53);
		  comm[owner].writeShort(trnB.id);
		  comm[owner].writeShort(trnB.bhulltype);
		  comm[owner].writeShort(trnB.bengtype);
		  comm[owner].writeShort(trnB.bbeamtype);
		  comm[owner].writeShort(trnB.bbeams);
		  comm[owner].writeShort(trnB.btorptype);
		  comm[owner].writeShort(trnB.blaunchs);
		  comm[owner].writeShort(0); /* What's this? Not fighterbays */
	      }
	      if (rstB.getTorpTech() != trnB.getTorpTech()) {
		  // NewCommand(6,54);
		  ptr[owner][commCount[owner]] = comm[owner].getPosition();
		  commCount[owner] += 1;
		  comm[owner].writeShort(54);
		  comm[owner].writeShort(trnB.id);
		  comm[owner].writeShort(trnB.getTorpTech());
	      }
	  }
      }

      /* other commands */
      for (int j=1; j<=races; j++) {
	  if (trnUniv.getGen(j) != null) {
	      Gen gen = trnUniv.getGen(j);
	      for (int i=0; i<trnUniv.getMessageCount(); i++) {
		  Message mess = trnUniv.getMessage(i);
		  // NewCommand(newmess.mess[i].messlen+8,60);
		  if (mess.from == j) {
		    ptr[j][commCount[j]] = comm[j].getPosition();
		    commCount[j] += 1;
		    comm[j].writeShort(60);
		    comm[j].writeShort(mess.content.length());
		    comm[j].writeShort(j); // from who could a message be?
		    comm[j].writeShort(mess.to);
		    mess.writeMessageTo(comm[j]);
		  }
	      };
	      if (gen.hasnewpwd) {
		  ptr[j][commCount[j]] = comm[j].getPosition();
		  commCount[j] += 1;
		  System.out.println("New password found");
		  // NewCommand(14,61);
		  ptr[j][commCount[j]] = comm[j].getPosition();
		  commCount[j] += 1;
		  comm[j].writeShort(61);
		  comm[j].writeShort(0);
		  comm[j].writeString(gen.newpwd);
	      }
	  }
      }

      /* Finish up */
      for (int j=1; j<=races; j++) {
	if (trnUniv.getGen(j) != null) {  // isPlaying()...
	  OutputStream outfile = ofo.open("player"+j+".trn");
	  trnFile[j] = new LSBWriter(outfile);
	  trnFile[j].writeShort(j);
	  trnFile[j].writeInt(commCount[j]);
	  trnFile[j].resetPartial();
	  trnFile[j].writeString(trnUniv.gen[j].timestamp);
	  csAddition[j] = 13 + 3*trnFile[j].getAndResetPartial(); // nasty, Tim
	  trnFile[j].writeShort(0); // sometimes 0, seems meaningless
	  trnFile[j].writeShort(trnUniv.gen[j].hostid);

	  if (commCount[j] != 0) {
	      trnFile[j].writeByte(0); // useless zero byte when >0 commands, Tim?
	      for (int k=0; k<commCount[j]; k++) {
		  // header, useless zero, ptrs, BASIC offset
		  trnFile[j].writeInt(ptr[j][k] + 28+1+4*commCount[j]+1);
	      }
	      comm[j].flush(); // otherwise still buffered and not available :(
	      trnFile[j].writeByteArray(((ByteArrayOutputStream)comm[j].getOutputStream()).toByteArray());
	  }

	  cs[j] = trnFile[j].getChecksum()+csAddition[j];
	  trnFile[j].writeInt(cs[j]);
	}
      }

      for (int j=1; j<=races; j++) {
	  if (trnUniv.getGen(j) != null) {
	  trnFile[j].writeString("Kero"); // personal touch
	  trnFile[j].writeByteArray(fizz); // registration
	  // checksums of all races found, Tim uses this to detect registration
	  // breaches.
	  for (int k=1; k<=races; k++) {
	      trnFile[j].writeInt( (trnFile[k] == null) ? 0 : cs[k] );
	  }
	  trnFile[j].close();
	  Generator.copyFileToDir(workdir.toString()+"player"+j+".trn", workdir.toString()+trnUniv.getTurnNumber());
	}
      }
      } catch (IOException ioe) {
	  crash(11, "Could not finish some TRN");
      }
  }

  /**
   * Put a Universe into rst[] or trn[], reallocating the arrays if necessary.
   */
  private void putUniverse(Universe u,int turnnr) {
    int when=u.get();
    // report("Universe from turn "+turnnr+" ("+(when==RstOrTrn.RST?"RST/dis":"TRN/dat")+")");
    switch (when) {
    case RstOrTrn.RST:
      if (rst.size()<turnnr+1) {
	rst.setSize(turnnr+1);
      }
      rst.setElementAt(u,turnnr);
      break;
    case RstOrTrn.TRN:
      if (trn.size()<turnnr+1) {
	trn.setSize(turnnr+1);
      }
      trn.setElementAt(u,turnnr);
      break;
    default:
      break;
    }
    if (turnnr>currentTurn) currentTurn=turnnr;
  }

  public int getCurrentTurn() {
    return currentTurn;
  }

  public Universe getUniverse() {
    return getUniverse(currentTurn,RstOrTrn.TRN);
  }

  public Universe getUniverse(int turnnr) {
    return getUniverse(turnnr, RstOrTrn.TRN);
  }

  protected Universe getNoHistoryUniverse(int turnnr,int when) {
    switch (when) {
    case RstOrTrn.RST:
      // System.err.println("getNoHistoryUniverse("+turnnr+","+when+":RST)");
      if (rst.size()<=turnnr) return null;
      return (Universe)rst.elementAt(turnnr);
      // break;
    case RstOrTrn.TRN:
      // System.err.println("getNoHistoryUniverse("+turnnr+","+when+":TRN)");
      if (trn.size()<=turnnr) return null;
      return (Universe)trn.elementAt(turnnr);
      // break;
    default:
      return null;
      // break;
    }
  }

  protected Universe getNoHistoryUniverse(String timestamp,int when) {
    Vector u;
    if (when==RstOrTrn.RST) u=rst; else u=trn;
    for (int i=1;i<=currentTurn;i++) {
      if ((u.size()>=i+1) && (u.elementAt(i)!=null)) {
	Universe universe=(Universe)u.elementAt(i);
	if (universe.getTimestamp().equals(timestamp)) return universe;
      }
    }
    return null;
  }

  public Universe getUniverse(int turnnr, int when) {
    switch (when) {
    case RstOrTrn.RST:
      // System.err.println("getUniverse("+turnnr+","+when+":RST)");
      if (rst.size()<=turnnr) rst.setSize(turnnr+1);
      if (rst.elementAt(turnnr)==null) updateFromHistory(turnnr);
      return (Universe)rst.elementAt(turnnr);
      // break;
    case RstOrTrn.TRN:
      // System.err.println("getUniverse("+turnnr+","+when+":TRN)");
      if (trn.size()<=turnnr) trn.setSize(turnnr+1);
      if (trn.elementAt(turnnr)==null) updateFromHistory(turnnr);
      if (trn.elementAt(turnnr)==null) {
	Universe ru=getUniverse(turnnr,RstOrTrn.RST);
	if (ru!=null) {
	  Universe tu=ru.toTrnUniverse();
	  putUniverse(tu,turnnr);
	}
      }
      return (Universe)trn.elementAt(turnnr);
      // break;
    default:
      return null;
      // break;
    }
  }

  public Universe getPreviousUniverse(int turnnr,int whence) {
    Universe u=null;
    while (u==null) {
      turnnr--;
      if (turnnr==0) turnnr=getCurrentTurn();
      u=getUniverse(turnnr,whence);
    }
    return u;
  }

  public Universe getNextUniverse(int turnnr,int whence) {
    // System.err.println("getNextUniverse("+turnnr+","+whence+")");
    Universe u=null;
    while (u==null) {
      turnnr++;
      u=getUniverse(turnnr,whence);
      if (u==null) {
	System.err.println("Universe "+turnnr+" not found");
	if (turnnr>getCurrentTurn()) turnnr=0;
      }
    }
    return u;
  }

  /**
   * Host current universe into a new universe
   */
  public Universe getHostedNextUniverse(int turnnr) {
    // System.err.println("getHostedNextUniverse("+turnnr+")");
    // Hosting is only allowed for 'future' turns
    if (turnnr >= getCurrentTurn()) {
      Universe hosted = getUniverse(turnnr, RstOrTrn.TRN).deepCopy();
      hosted.hostSelf();
      putUniverse(hosted, turnnr + 1);
    };
    return getNextUniverse(turnnr, RstOrTrn.RST);
  }

  public void checkConsistency() {
    Universe trnuniv = getUniverse(getCurrentTurn(), RstOrTrn.TRN);
    Universe rstuniv = getUniverse(getCurrentTurn(), RstOrTrn.RST);
    Enumeration enum = trnuniv.getLocations();
    while (enum.hasMoreElements()) {
      Location loc = (Location)enum.nextElement();
      loc.assertConsistency();
    }
    report("Consistency Check passed.");
  }

    /**
     * @param type Hull type from 1 to 105.
     */
    public Hull getHull(int type) {
	return hull[type];
    }

    public Truehull getTrueHull(int player) {
	return truehull[player];
    }

    /**
     * @param hulltype Hull type for player (from 1 to 20).
     */
    public int getTrueHullIndex(int player, int bhulltype) {
	return getTrueHull(player).build[bhulltype];
    }

    /**
     * @param hulltype Hull type for player (from 1 to 20).
     */
    public Hull getHull(int player, int hulltype) {
	int race = getUniverse().getConfig().getPlayerRace(player);
	int index = getTrueHullIndex(race, hulltype);
	if (index == 0) return null;
	return getHull(index);
    }

    public Hullfunc getHullfunc() {
	return hullfunc;
    }

  /**
   * Retrieve engine of the proper type.
   * @param type Corresponds with the number Ship.engtype
   */
  public Engine getEngine(int type) {
    if (type==UNKNOWN) return null;
    if (type<0) return null; // illegal type (exception?)
    if (type>engines) return null; // illegal type (exception?)
    if (type==0) return null; // non-existing engine
    return engine[type];
  }

  /**
   * Retrieve torpedo of the proper type.
   * @param type Corresponds with the number Ship.torptype
   */
  public Torpedo getTorpedo(int type) {
    if (type==UNKNOWN) return null;
    if (type<0) return null; // illegal type (exception?)
    if (type>torpedoes) return null; // illegal type (exception?)
    if (type==0) return null; // non-existing torpedo
    return torpedo[type];
  }

  /**
   * Retrieve beam of the proper type.
   * @param type Corresponds with the number Ship.beamtype
   */
  public Beam getBeam(int type) {
    if (type==UNKNOWN) return null;
    if (type<0) return null; // illegal type (exception?)
    if (type>beams) return null; // illegal type (exception?)
    if (type==0) return null; // non-existing beam
    return beam[type];
  }

  public InputFileOpener getIFO() {
    return ifo;
  }

  public OutputFileOpener getOFO() {
    return ofo;
  }

    // the word editable is incorrect.
    // waitForEditable on an RST-Universe will block forever?
    
    /**
     * Should be used as:
     *   if (game.couldGrabLock()) {
     *     // do the important stuff
     *     unlock();
     *   }
     *  Typical use for regular, single-thread client-stuff.
     */
    public synchronized void unlock() {
	locked = false;
	// System.err.println("Lock given free");
	notifyAll();
    }

    public synchronized boolean couldGrabLock() {
	if (!locked) {
	    locked = true;
	    // System.err.println("Lock obtained");
	    return true;
	}
	return false;
    }

    /**
     * Should be used if lock is to be obtained.
     * For example, the blitz connection needs the lock to make the TRN.
     * This is a second thread, but non-editing. Multiple editing threads
     * should use this method as well, like:
     *   waitForLocked();
     *   // do the imporant stuff
     *   resetLocked();
     */
    public synchronized void waitForLock() {
	while (locked) {
	    try { wait(); } catch (InterruptedException ie) {}
	}
	locked = true;
    }
}
