// File: LSBReader.java

package jvpc.vga;

import java.io.*;

/**
 * Reads Least Significant Byte First (LSB) numbers and Strings from an <CODE>
 * InputStream</code>.
 *
 * @author    Kero van Gelder
 * @copyright Kero van Gelder (the GNU General Public License applies)
 * @since     10 VII 1998
 * @version   7 III 1998
 */
public class LSBReader extends Reader {
  private InputStream is;
  private int pos=0;

  public LSBReader(InputStream is) {
    this.is=is;
  }

  public int available() throws IOException {
    return is.available();
  }

  public byte readByte() throws IOException {
    pos++;
    return (byte)is.read();
  }

  /**
   * Reads an unsigned byte.
   * @return The unsigned byte <EM>as an int</em>.
   */
  public int readUnsignedByte() throws IOException {
    pos++;
    return is.read();
  }
  
  /**
   * Reads a <CODE>short</code> in an LSB fashion.
   */
  public short readShort() throws IOException {
    int one,two;
    one=is.read(); // simulates unsigned byte behavior;
    two=is.read();
    pos+=2;
    // add, then cast and return.
    return (short)((two<<8)+one);
  }
  
  /**
   * Reads an unsigned <CODE>short</code> in an LSB fashion.
   * @return The unsigned short <EM>as an int</em>.
   */
  public int readUnsignedShort() throws IOException {
    int one,two;
    one=is.read(); // simulates unsigned byte behavior;
    two=is.read();
    pos+=2;
    // add, then return (casting not necessary).
    return (two<<8)+one;
  }

  /**
   * Reads an <CODE>int</code> in an LSB fashion.
   */
  public int readInt() throws IOException {
    int one,two,three,four;
    one=is.read(); // simulates unsigned byte behavior
    two=is.read();
    three=is.read();
    four=is.read();
    pos+=4;
    // add, then return (casting not necessary).
    return (four<<24)+(three<<16)+(two<<8)+one;
  }

  /**
   * Reads an unsigned <CODE>int</code> in an LSB fashion.
   * @return The unsigned int <EM>as a long</em>.
   */
  public long readUnsignedInt() throws IOException {
    int one,two,three,four;
    one=is.read(); // simulates unsigned byte behavior
    two=is.read();
    three=is.read();
    four=is.read();
    pos+=4;
    // cast, then add and return.
    return ((long)(four)<<24)+(long)((three)<<16)+((long)(two)<<8)+(long)(one);
  }

  /**
   * Reads a <CODE>long</code> in an LSB fashion.
   * @exception @see java.io.InputStream.read()
   */
  private long readLong() throws IOException {
    int one,two,three,four,five,six,seven,eight;
    one=is.read(); // simulates unsigned byte behavior
    two=is.read();
    three=is.read();
    four=is.read();
    five=is.read();
    six=is.read();
    seven=is.read();
    eight=is.read();
    pos+=8;
    // cast, then add and return.
    return ((long)(four)<<24)+((long)(three)<<16)+((long)(two)<<8)+(long)(one)+
      (long)(eight)<<56+(long)(seven)<<48+(long)(six)<<40+(long)(five)<<32;
  }

  /**
   * Reads a String, converts bytes to chars.
   *
   * @param size Size of the String
   */
  public StringBuffer readString(int size) throws IOException {
    StringBuffer result=new StringBuffer(size); // sets capacity
    result.setLength(size); // sets actual length (length<=capacity)
    int i,ch;
    for (i=0;i<size;i++) {
      // read ith byte, convert it to UNIcode char
      ch=is.read();
      result.setCharAt(i,(char)ch); // plain ASCII -> conversion succesfull
    }
    pos+=size;
    return result;
  }

    public byte[] readByteArray(int size) throws IOException {
	byte[] result = new byte[size];
	int i,ch;
	for (i=0; i<size; i++)  result[i] = (byte)is.read();
	pos += size;
	return result;
    }

  public int read(char[] string, int size, int amount) throws IOException {
    throw new IOException("Not implemented");
  }

  public void close() throws IOException {
    is.close();
  }

  /**
   * Skips a number of bytes.
   *
   * @param size Amount of skipped bytes.
   */
  public long skip(long size) throws IOException {
    pos+=size;
    long result = is.skip(size);
    return result;
  }

  /**
   * Skip until the given position.
   * @param position Given position.
   */
  public void skipUntil(int position) throws IOException {
    if (pos == position) return;
    if (pos<position) {
      //System.err.println("trying to skip "+(position-pos)+" bytes");
      long result = is.skip(position-pos);
      //System.err.println("did skip "+result);
      pos=position;
    } else {
      System.err.println("Can not go back from "+pos+" to "+position+"!");
    }
  }

  public int getPosition() {
    return pos;
  }

    public static void main(String arg[]) {
	System.err.println("Min char value: "+(int)Character.MIN_VALUE);
	System.err.println("Max char value: "+(int)Character.MAX_VALUE);
	System.err.println("int "+65+" is char "+(char)65);
	System.err.println("int "+200+" is char "+(char)200);
	System.err.println("int "+20102+" is char "+(char)20102);
	System.err.println("int "+123456+" is char "+(char)123456);
	System.err.println("int "+200+" is byte "+(byte)200+" is char-via-byte "+(char)(byte)200);
    }
}
