// File: Location.java

package jvpc.vga;

import java.util.*;
import kero.util.*;

/**
 * Provides information on a specified location in an universe.
 * Can perform consistency check for the Location, which is the best internal
 * test running in this program.
 *
 * @author Kero van Gelder
 * @copyright Kero van Gelder (the GNU General Public License applies)
 * @since January 1999 ?
 * @version 14 III 2001
 */
public class Location
    extends Located
{
    Planet planet = null;	       // Planet (PData, BData) at this location
    Vector ships = new Vector(); // All ships on this location
    Vector minefields = new Vector(); // All minefields on this location
    Vector explosions = new Vector(); // All explosions on this location
    Vector ionstorms = new Vector(); // All ionstorms on this location
    Vector wormholes = new Vector(); // All wormholes on this location
    Planet well = null;	       // Planet which owns the planet well on this loc
    // for the enumeration
    Vector otherLocateds = new Vector(); // e.g.  jvpc.extdata.Marker

    public BData getBase() {
	if (planet instanceof BData) return (BData)planet;
	return null;
    }

    public Planet getPlanet() {
	return planet;
    }

    public Vector getShips() {
	return ships;
    }

    public Ship getShip(int index) {
	return (Ship)ships.elementAt(index);
    }

    public Vector getMinefields() {
	return minefields;
    }

    public Vector getExplosions() {
	return explosions;
    }

    public Vector getIonstorms() { return ionstorms; }

    public Vector getWormholes() {
	return wormholes;
    }

    public Planet getWell() {
	return well;
    }

    public Vector getOthers() {
	return otherLocateds;
    }

    public int shipsHissing() {
	if (!getConfig().allowHiss()) return 0;
	int hissing = 0;
	Enumeration enum = ships.elements();
	while (enum.hasMoreElements()) {
	    Ship ship = (Ship)enum.nextElement();
	    if ((ship.getOwner() == Race.LIZARDS) && !ship.isFreighter()) {
		if (ship.getMission() == Mission.SPECIAL) hissing++;
	    }
	}
	int maxHissing = getConfig().maxShipsHissing();
	if (hissing > maxHissing) hissing = maxHissing;
	return hissing;
    }

    public Location(Universe u, int x, int y) {
	super(u, x, y);
    }

    /**
     * Look for all objects in the universe at this location,
     * and add it to this location.
     *
     * This takes O(n), with n the # of objects in the universe
     */
    public void update() {
	Enumeration enum=universe.getLocatedObjects();
	while (enum.hasMoreElements()) {
	    putLocated((Located)enum.nextElement());
	}
    }

    public void putLocated(Located loc) {
	if (!loc.liesWithin(getPoint())) return;

	if (loc instanceof Planet) {
	    planet = (Planet)loc;
	} else if (loc instanceof Ship) {
	    Ship s = (Ship)loc;
	    int i = 0;
	    while ((i < ships.size()) &&
		   (((Ship)ships.elementAt(i)).getId() < s.getId())) {
		i++;
	    }
	    ships.insertElementAt(loc, i);
	} else if (loc instanceof Ionstorm) {
	    ionstorms.addElement(loc);
	} else if (loc instanceof Wormhole) {
	    wormholes.addElement(loc);
	} else if (loc instanceof Minefield) {
	    if (minefields.indexOf(loc) < 0) minefields.addElement(loc);
	} else if (loc instanceof Explosion) {
	    if (explosions.indexOf(loc) < 0) explosions.addElement(loc);
	} else if (loc instanceof Vcr) {
	    // ignore for the moment
	} else {
	    report("what do we have here? "+loc);
	    otherLocateds.addElement(loc);
	}
    }

    public boolean contains(Located loc) {
	Enumeration enum = locateds(true, true, true);
	while (enum.hasMoreElements()) {
	    Located enumLoc = (Located)enum.nextElement();
	    if (enumLoc == loc) return true;
	}
	return false;
    }

    /** Get an element from a complete list (performing selection atm) */
    public Located getElement(int index) {
	// if (index == -1) return location.getPlanet();
        Enumeration enum = locateds(true, true, true);
	while (enum.hasMoreElements()) {
	    Located loc = (Located)enum.nextElement();
	    if (index == 0) return loc;
	    index--;
	}
	return null;
    }

    /** 
     * Enumeration of this location;
     * @return this returns itself, cause this ain't a beautiful list.
     */
    public Enumeration locateds(boolean usePlanet, boolean useShips,
				boolean useMinefields) {
      return new LocationEnumeration(this, usePlanet, useShips, useMinefields);
    }

    public String toString() {
	return "Location(x="+x+",y="+y+
	    ",#p="+(planet!=null?1:0)+
	    ",#s="+ships.size()+
	    ",#m="+minefields.size()+
	    ",#e="+explosions.size()+
	    ",#i="+ionstorms.size()+
	    ",#w="+wormholes.size()+")";
    }

    public Location getRstLocation() {
	Assert.assert(!isRst());
	Universe rstuniverse = getGame().getUniverse(universe.getTurnNumber(), RstOrTrn.RST);
	Location result = rstuniverse.getNearestTo(x, y);
	if ((result.getX() != x) || (result.getY() != y)) result = null;
	return result;
    }

    public int getSupplies() {
	int result = 0;
	Planet planet = getPlanet();
	if (planet != null) result += planet.getSupplies();
	Enumeration enum = ships.elements();
	while (enum.hasMoreElements()) {
	    Ship ship = (Ship)enum.nextElement();
	    result += ship.getSupplies();
	}
	return result;
    }

    public int getFighters() {
	int result = 0;
	BData base = getBase();
	if (base != null) {
	    result += base.getFighters();
	}
	Enumeration enum = ships.elements();
	while (enum.hasMoreElements()) {
	    Ship ship = (Ship)enum.nextElement();
	    result += ship.getFighters();
	}
	return result;
    }

    public int getTorpedoes(int torptype) {
	int result = 0;
	BData base = getBase();
	if (base != null) {
	    result += base.getTorpedoes(torptype);
	}
	Enumeration enum = ships.elements();
	while (enum.hasMoreElements()) {
	    Ship ship = (Ship)enum.nextElement();
	    if (ship.getTorpedoType() == torptype) result += ship.getTorpedoes();
	}
	return result;
    }

  /**
   * Check consistency on anything that is Our()s. We don't care about other
   * peoples planets and ships, because we won't produce TRN's for them.
   *
   * <P>The check is complete within reason (but can never be perfect).
   *
   * @return A string that tells what is not consistent. Hence, returning null
   * means everything is OK.
   */
  public String checkConsistency() {
    // WARNING: If you do not understand all VGA-Planets data-structures or the
    // entire jvpc.vga package, DO NOT ALTER THIS FUNCTION. [Kero]
    System.err.print("C");
    if (isRst()) return "Not a TRN Location";
    Location rstloc = getRstLocation();
    if (rstloc == null) return "No equivalent RST Location";
    if ((rstloc.getX() != getX()) && (rstloc.getY() != getY())) {
	assert(planet == null);
	assert(ships.isEmpty());
	return null;
    }

    int rstneutr=0, trnneutr=0;
    int rsttrit=0, trntrit=0, rstdur=0, trndur=0, rstmolyb=0, trnmolyb=0;
    int rstsupp=0, trnsupp=0;
    int rstmc=0, trnmc=0;
    int rstcolclans=0, trncolclans=0;

    if (rstloc.planet != null) {
      if (planet == null) return "planet == null";
      if (rstloc.planet.getId() != planet.getId()) return "planet Ids unequal";
      if (rstloc.planet.isOurPlanet()) {
	if (!planet.isOurPlanet()) return "!planet.isOurPlanet";
	PData rstpdata = (PData)rstloc.planet;
	PData trnpdata = (PData)planet;
	rstneutr += rstpdata.getNeutronium();
	trnneutr += trnpdata.getNeutronium();
	rsttrit += rstpdata.getTritanium();
	trntrit += trnpdata.getTritanium();
	rstdur += rstpdata.getDuranium();
	trndur += trnpdata.getDuranium();
	rstmolyb += rstpdata.getMolybdenium();
	trnmolyb += trnpdata.getMolybdenium();
	rstsupp += rstpdata.getSupplies();
	trnsupp += trnpdata.getSupplies();
	rstmc += rstpdata.getMC();
	trnmc += trnpdata.getMC();
	//System.err.println("P"+planet.getId()+": "+rstsupp+"/"+trnsupp);
	if (rstpdata.struct.length != trnpdata.struct.length) return "struct.length mismatch";
	for (int i=0; i<rstpdata.struct.length; i++) {
	  /* Check amount of structures has not decreased. Fighters & Torps are
	   * also CargoTransferable, so need another check.
	   */
	  if (i == Build.FIGHTER) {
	    if (getFighters() < rstloc.getFighters())
		return "#fighters down from "+rstloc.getFighters()+" to "+getFighters();
	  } else if (i > Build.TORPOFFSET && i <= Build.TORPOFFSET+torpedoes) {
	    int torptype = i-Build.TORPOFFSET;
	    if (getTorpedoes(torptype) < rstloc.getTorpedoes(torptype)) return "#torps, type "+torptype;
	  } else {
	    if (trnpdata.struct[i] < rstpdata.struct[i]) return "lesser struct["+i+"]";
	  }
	  rsttrit += rstpdata.struct[i] * rstpdata.needTrit(i);
	  trntrit += trnpdata.struct[i] * trnpdata.needTrit(i);
	  rstdur += rstpdata.struct[i] * rstpdata.needDur(i);
	  trndur += trnpdata.struct[i] * trnpdata.needDur(i);
	  rstmolyb += rstpdata.struct[i] * rstpdata.needMolyb(i);
	  trnmolyb += trnpdata.struct[i] * trnpdata.needMolyb(i);
	  rstsupp += rstpdata.struct[i] * rstpdata.needSupp(i);
	  trnsupp += trnpdata.struct[i] * trnpdata.needSupp(i);
	  rstmc += rstpdata.struct[i] * rstpdata.needMC(i);
	  trnmc += trnpdata.struct[i] * trnpdata.needMC(i);
	  //System.err.print(" "+rstsupp+"/"+trnsupp);
	}
	//System.err.println();
	if (rstpdata.buildingBase()) return "building base in result?";
	if (trnpdata.hasBase() && trnpdata.buildingBase()) return "two bases?";
	rstcolclans += rstpdata.getColClans();
	trncolclans += trnpdata.getColClans();
      } else {
	if (planet.isOurPlanet()) return "planet.isOurPlanet";
      }
    } else {
      if (planet != null) return "planet != null";
    }

    // Base checking
    BData rstbase = rstloc.getBase();
    BData trnbase = getBase();
    if (rstbase != null) {
      if (trnbase == null) return "base == null";
      if (rstbase.getId() != trnbase.getId()) return "base Ids unequal";
      if (rstbase.getOwner() != trnbase.getOwner()) return "base owners";
      for (int i=0; i<nrOfTechLevels; i++) {
	  if (trnbase.techLevel[i] < rstbase.techLevel[i]) {
	      return "techLevel["+i+"] lowered from "+rstbase.techLevel[i]+
		  " to "+trnbase.techLevel[i];
	  }
	  rstmc += totalTechLevelCost[rstbase.techLevel[i]];
	  trnmc += totalTechLevelCost[trnbase.techLevel[i]];
      }
      // torps, fighters, defense and ship parts done under planet (struct[])
    } else {
      if (trnbase != null) return "base != null";
    }

    // Ships are in the ship-Vectors in the same order, usually :-)
    Enumeration rstenum = rstloc.ships.elements();
    Enumeration trnenum = ships.elements();
    while (rstenum.hasMoreElements()) {
      if (!trnenum.hasMoreElements()) return "Out of trn elements";
      Ship rstship, trnship;
      rstship = (Ship)rstenum.nextElement();
      trnship = (Ship)trnenum.nextElement();
      // if (rstship.getId() != trnship.getId()) return "ship Id's unequal (rst: "+rstship.getId()+" trn: "+trnship.getId()+")";
      if (rstship.isOurShip()) {
	if (!trnship.isOurShip()) return "!trnship.isOurShip";
	rstneutr += rstship.getNeutronium() + rstship.toplanet.getNeutronium() + rstship.toship.getNeutronium();
	trnneutr += trnship.getNeutronium() + trnship.toplanet.getNeutronium() + trnship.toship.getNeutronium();
	rsttrit += rstship.getTritanium() + rstship.toplanet.getTritanium() + rstship.toship.getTritanium();
	trntrit += trnship.getTritanium() + trnship.toplanet.getTritanium() + trnship.toship.getTritanium();
	rstdur += rstship.getDuranium() + rstship.toplanet.getDuranium() + rstship.toship.getDuranium();
	trndur += trnship.getDuranium() + trnship.toplanet.getDuranium() + trnship.toship.getDuranium();
	rstmolyb += rstship.getMolybdenium() + rstship.toplanet.getMolybdenium() + rstship.toship.getMolybdenium();
	trnmolyb += trnship.getMolybdenium() + trnship.toplanet.getMolybdenium() + trnship.toship.getMolybdenium();
	rstsupp += rstship.getSupplies() + rstship.toplanet.getSupplies() + rstship.toship.getSupplies();
	trnsupp += trnship.getSupplies() + trnship.toplanet.getSupplies() + trnship.toship.getSupplies();
	rstmc += rstship.getMC();
	trnmc += trnship.getMC();
	rstcolclans += rstship.getColClans() + rstship.toplanet.getColClans() + rstship.toship.getColClans();
	trncolclans += trnship.getColClans() + trnship.toplanet.getColClans() + trnship.toship.getColClans();
	if (rstship.bays > 0) {
	  if (trnship.bays != rstship.bays) return "fighter bay mismatch";
	  // we _know_ there is only TM$ involved for fighters
	  int rstfighters = rstship.cargo[Cargo.AMMUNITION].amount;
	  int trnfighters = trnship.cargo[Cargo.AMMUNITION].amount;
	  rsttrit += rstfighters * Build.needtrit[Build.FIGHTER];
	  trntrit += trnfighters * Build.needtrit[Build.FIGHTER];
	  rstmolyb += rstfighters * Build.needmolyb[Build.FIGHTER];
	  trnmolyb += trnfighters * Build.needmolyb[Build.FIGHTER];
	  rstmc += rstfighters * Build.needmc[Build.FIGHTER];
	  trnmc += trnfighters * Build.needmc[Build.FIGHTER];
	}
	if (rstship.launchs > 0) {
	  if (trnship.launchs != rstship.launchs)
	      return "torp launcher mismatch";
	  if (trnship.getTorpedoType() != rstship.getTorpedoType())
	      return "torp type mismatch";
	  //if (rstship.getTorpedoType() == 0) {
	      // oh yes, I managed to do this with latest and greatest 3.22.47
	  // } else {
	  if ((rstship.getTorpedoType() > 0) &&
		     (rstship.getTorpedoType() <= torpedoes)) {
	      // there is only TDM$ involved in torpedoes; 1 TDM and x $
	      int rsttorps = rstship.cargo[Cargo.AMMUNITION].amount;
	      int trntorps = trnship.cargo[Cargo.AMMUNITION].amount;
	      rsttrit += rsttorps;
	      trntrit += trntorps;
	      rstdur += rsttorps;
	      trndur += trntorps;
	      rstmolyb += rsttorps;
	      trnmolyb += trntorps;
	      Game game = universe.getGame();
	      rstmc += rsttorps * game.torpedo[rstship.getTorpedoType()].unitmc;
	      trnmc += trntorps * game.torpedo[trnship.getTorpedoType()].unitmc;
	  } else {
	      return "Illegal torptype!";
	  }
	}
      } else {
	if (trnship.isOurShip()) return "trnship.isOurShip";
      }
    }
    if (trnenum.hasMoreElements()) return "Out of rst elements";

    if (rstneutr != trnneutr) return "Unequal neutr "+rstneutr+" "+trnneutr;
    if (rsttrit != trntrit) return "Unequal of trit";
    if (rstdur != trndur) return "Unequal dur";
    if (rstmolyb != trnmolyb) return "Change of "+(trnmolyb-rstmolyb)+" molyb";
    if (rstsupp < trnsupp) return "Raise of supplies "+rstsupp+"->"+trnsupp;
    if (rstsupp+rstmc != trnsupp+trnmc) {
	warn("S/$: "+rstsupp+"/"+rstmc+" -> "+trnsupp+"/"+trnmc);
	return "Unequal MC";
    }
    if (rstcolclans != trncolclans) return "ColClans MisMatch "+rstcolclans+" "+trncolclans;
    return null;
  }

    public void assertConsistency() {
	String check = checkConsistency();
	if (check != null) {
	    System.err.println(check);
	    Location rstLoc = getRstLocation();
	    System.err.println("rst: "+rstLoc);
	    Enumeration enum = rstLoc.getShips().elements();
	    while (enum.hasMoreElements()) {
		Ship ship = (Ship)enum.nextElement();
		System.err.println(ship.getReport());
	    }

	    System.err.println("trn: "+this);
	    enum = getShips().elements();
	    while (enum.hasMoreElements()) {
		Ship ship = (Ship)enum.nextElement();
		System.err.println(ship.getReport());
	    }

	    crash(1002, check);
	}
    }
}


class LocationEnumeration implements Enumeration {
    Vector vector = new Vector(); // Vector of Vectors, even
    /** the indices to be used by the next call to nextElement() */
    int vecCount, subCount;

    public LocationEnumeration(Location loc, boolean usePlanet,
			       boolean useShips, boolean useMinefields) {
	int vecCount = 0;
	if (usePlanet && (loc.getPlanet() != null)) {
	    Vector planetVector = new Vector();
	    planetVector.addElement(loc.getPlanet());
	    vector.addElement(planetVector);
	}
	if (useShips) vector.addElement(loc.getShips());
	if (useMinefields) vector.addElement(loc.getMinefields());
	vector.addElement(loc.getIonstorms());
	vector.addElement(loc.getWormholes());
	vector.addElement(loc.getExplosions());
	vector.addElement(loc.getOthers());
	vecCount = subCount = 0;
    }

    public boolean hasMoreElements() {
	while (vecCount < vector.size()) {
	    Vector v = (Vector)vector.elementAt(vecCount);
	    if (subCount < v.size()) return true;
	    vecCount++;
	    subCount = 0;
	}
	return false;
    }

    public Object nextElement() {
	Vector v = (Vector)vector.elementAt(vecCount);
	Object result = v.elementAt(subCount);
	subCount++;
	return result;
    }
}
