// File Message.java

package jvpc.vga;

import java.io.*;

/**
 * Provides all fields for a message for VGA-Planets.
 * A message is max 15 lines of max 40 chars (plus a carriage-return per line).
 * VGA-Planets is made by Tim Wisseman.
 *
 * @author    Kero van Gelder
 * @copyright Kero van Gelder (the GNU General Public License applies)
 * @since     12 II 1999
 * @version   15 II 1999
 */
public class Message extends RstOrTrn {
  int from,to;
  StringBuffer content;

  public Message(Universe u) {
    super(u);
    from=to=UNKNOWN;
    content=new StringBuffer();
  }

  /**
   * read one MData-message, that is an incoming message.
   */
  public void readMDataFrom(LSBReader source,int length,Universe universe, int player) throws IOException {
    // from/Sender is not in the message; why not, Tim?
    to=player;
    content.setLength(length);
    // System.err.println("MData: to "+player+", length "+length);
    char c=(char)(source.readUnsignedByte()-MESSCODE);
    int i=0,j=0;
    while (j<length) {
      // System.err.print("("+i+","+j+") ");
      switch (c) {
      case '\012': // skip possible following '\015', as Tim does
	content.setCharAt(i,c);
	if (j<length-1) {
	  c=(char)(source.readUnsignedByte()-MESSCODE);
	  if ((c=='\015') && (j<length-2)) { // yes, length-2 here
	    c=(char)(source.readUnsignedByte()-MESSCODE);
	    j++;
	  }
	}
	j++;
	break;
      case '\015':
	c='\012';
	// no break here
      default:
	content.setCharAt(i,c);
	if (j<length-1) {
	  c=(char)(source.readUnsignedByte()-MESSCODE);
	}
	j++;
      } // end switch
      i++;
    }
    content.setLength(i);

    String str = content.toString();
    if (str.startsWith("(-i0")) {
      // ionstorm
      parseIonstorm(str, universe);
    } else if (str.startsWith("(-m0")) {
      // minefield
      parseMinefield(str, universe, player);
    }
  }

  public void parseIonstorm(String str, Universe universe) throws IOException {
    int id = Integer.parseInt(str.substring(4, 7));
    Ionstorm is;
    if (universe.containsIonstorm(id)) {
      is = universe.getIonstorm(id);
    } else {
      is = new Ionstorm(universe, id);
      universe.putIonstorm(is);
    }
    /* Apply a simple word-followed-by-number scheme; this can be done since
       there are no funny names in the message tha could accidentally contain
       the same word */
    StreamTokenizer parser = new StreamTokenizer(new StringReader(str));
    int type;
    int state = 0;
    while ((type = parser.nextToken()) != StreamTokenizer.TT_EOF) {
      if (type == StreamTokenizer.TT_WORD) {
	if (parser.sval.equals("At")) {
	  state = 1;
	} else if (parser.sval.equals("Voltage")) {
	  state = 3;
	} else if (parser.sval.equals("Heading")) {
	  state = 4;
	} else if (parser.sval.equals("Warp")) {
	  state = 5;
	} else if (parser.sval.equals("Radius")) {
	  state = 6;
	}
      }
      if (type == StreamTokenizer.TT_NUMBER) {
	if (state == 1) {
	  is.setX((int)parser.nval);
	  state = 2;
	} else if (state == 2) {
	  is.setY((int)parser.nval);
	  state = 0;
	} else if (state == 3) {
	  is.voltage = (int)parser.nval;
	  state = 0;
	} else if (state == 4) {
	  is.heading = (int)parser.nval;
	  state = 0;
	} else if (state == 5) {
	  is.warp = (int)parser.nval;
	  state = 0;
	} else if (state == 6) {
	  is.radius = (int)parser.nval;
	  state = 0;
	}
      }
    }
  }

  public void parseMinefield(String str, Universe universe, int player) throws IOException {
    int id = Integer.parseInt(str.substring(4, 7));
    if (id == 0) return; // explosions, whatever. ignore.
    Minefield mf = universe.getMinefield(id);
    if (mf == null) {
      mf = new Minefield(universe, id);
      universe.putMinefield(mf);
    }
    // if (id == 22) System.out.println(str);
    /* Apply a simple word-followed-by-number scheme; this can be done since
       there are no funny names in the message tha could accidentally contain
       the same word */
    StreamTokenizer parser = new StreamTokenizer(new StringReader(str));
    int type;
    int state = 0;
    while ((type = parser.nextToken()) != StreamTokenizer.TT_EOF) {
      if (type == StreamTokenizer.TT_WORD) {
	if (state == 4) {
	  if (parser.sval.equals("They")) state = 5;
	  else state = 0;
	} else if (state == 5) {
	  if (parser.sval.equals("are")) state = 6;
	  else state = 0;
	} else if (state == 6) {
	  Player p = universe.getPlayerByAdjective(parser.sval);
	  mf.setOwner((p == null) ? 0 : p.player);
	  state = 0;
	} else if (parser.sval.equalsIgnoreCase("at")) {
	  state = 1;
	} else if (parser.sval.equals("contains")) {
	  state = 3;
	} else if (parser.sval.equals("weapons.")) {
	  state = 3;
	} else if (parser.sval.equals("destroyed!")) {
	  state = 3;
	} else {
	  state = 0;
	}
      }
      if (type == StreamTokenizer.TT_NUMBER) {
	if (state == 1) {
	  mf.setX((int)parser.nval);
	  state = 2;
	} else if (state == 2) {
	  mf.setY((int)parser.nval);
	  state = 4;
	} else if (state == 3) {
	  mf.setUnits((int)parser.nval);
	  state = 0;
	}
      }
    }
    if (mf.owner == UNKNOWN) mf.owner = player;
    // System.out.println("  ==>> "+mf.toString(true));
  }

  /**
   * read one Mess-message, that is an outgoing message.
   */
  public void readMessFrom(LSBReader source,int length,int player) throws IOException {
    content=new StringBuffer(length);
    content.setLength(length);
    for (int j=0;j<length;j++) {
      char c=(char)(source.readUnsignedByte()-MESSCODE);
      if (c=='\015') c='\012';
      content.setCharAt(j,c);
    }
  }

  public void readFromTrn(LSBReader source,int length,int player) throws IOException {
    from=source.readShort();
    to=source.readShort();
    content=new StringBuffer(length);
    content.setLength(length);
    for (int j=0;j<length;j++) {
      char c=(char)(source.readUnsignedByte()-MESSCODE);
      if (c=='\015') c='\012';
      content.setCharAt(j,c);
    }
  }

  public void setFromAndTo(int f,int t) {
    from=f;
    to=t;
  }


  /**
   * Writes one message-content, that is, somebody else must have
   * built the pointer index in this file before (including from/to
   * for outgoing messages
   */
  public void writeMessageTo(LSBWriter sink) throws IOException {
    int length=content.length();
    for (int j=0;j<length;j++) {
      char c=content.charAt(j);
      if (c=='\012') c='\015';
      sink.writeByte((byte)(c+MESSCODE));
    }
  }

  public int getSender() {
    return from;
  }

  public int getRecipient() {
    return to;
  }

  public int getLength() {
    return content.length();
  }
  
  public String getContent() {
    return content.toString();
  }

    public void setSender(int newfrom) {
	if (isEditable()) {
	    from = newfrom;
	    getGame().unlock();
	}
    }

    public void setRecipient(int newto) {
	if (isEditable()) {
	    to = newto;
	    getGame().unlock();
	}
    }

    /**
     * Sets contents of this Message to str. A simple check is performed to see
     * whether the message is too long. Also, it is not allowed to edit an
     * incoming message, of course.
     */
    public void setContent(String str) {
	if (str.length() > 40*25) return;
	if (isEditable()) {
	    content = new StringBuffer(str);
	    // throw some event...
	    getGame().unlock();
	}
    }
}
