// File: Minefield.java

package jvpc.vga;

import java.util.*;
import java.io.*;

/**
 * Contains a minefield for JVPC.
 *
 * @author    Kero van Gelder
 * @copyright Kero van Gelder (the GNU General Public License applies)
 * @since     13 I 1999
 */
public class Minefield
  extends Owned
  implements Intersection
{
  int units; // number of mines
  int isweb; // 1: is a webMinefield; 0: is a normal Minefield
  int fcodeplanet; // id of the planet which has the FC
  int scantype; // 0: lay; 1: sweep; 2: scan

  public Vector getTypes() {
    Vector res = super.getTypes();
    res.addElement("radius");
    res.addElement("web");
    res.addElement("id");
    res.addElement("units");
    return res;
  }

  public Vector getTypeMethods(String typeName) {
    Vector res = super.getTypeMethods(typeName);
    if ("id".equals(typeName)) {
      res.addElement("getFCPlanet");
    } else if ("radius".equals(typeName)) {
      res.addElement("getRadius");
    } else if ("web".equals(typeName)) {
      res.addElement("isWeb");
    } else if ("units".equals(typeName)) {
      res.addElement("getUnits");
    }
    return res;
  }

  public Minefield() {
      super(null);
  }

  public Minefield(Universe u) {
    super(u);
  }

  public Minefield(Universe u, int id) {
    super(u);
    this.id = id;
  }

  public Object clone() throws CloneNotSupportedException {
    return (Minefield)(super.clone());
  }

  public void readFromUtil(LSBReader source) throws IOException {
    // id already read in UtilGen
    x=source.readShort();
    y=source.readShort();
    owner=source.readShort();
    units=source.readInt();
    isweb=source.readShort();
    fcodeplanet=source.readShort();
    scantype=source.readShort();
  }

  public boolean isWeb() {
    return isweb==1;
  }

  public int getFCPlanet() {
    return fcodeplanet;
  }

  public int getUnits() {
    return units;
  }

  public void setUnits(int units) {
    this.units = units;
  }

  /**
   * Return # of mines after 1 turn decay
   */
  public int getUnitsNT() {
    int decay;
    if (isWeb()) {
      decay = getConfig().webMineDecayRate();
    } else {
      decay = getConfig().mineDecayRate();
    }
    return (int)(units * (100 - decay) / 100);
  }

  public double getRadius() {
    return Math.sqrt(units);
  }

  public Circle getCircle() {
    Circle c = new Circle(new MyPoint(getX(), getY()));
    c.setSqrRadius(units);
    return c;
  }

  /**
   * Returns true if point p lies on this object
   */
  public boolean liesOn(MyPoint p) {
    return getCircle().liesOn(p);
  }

  /**
   * Returns true if point p lies within this object
   */
  public boolean liesWithin(MyPoint p) {
    return getCircle().liesWithin(p);
  }

  /**
   * Returns true if line l lies (partly) within this object
   */
  public boolean liesWithin(Line l) {
    return getCircle().liesWithin(l);
  }

  /**
   * Returns true if rectangle r lies (partly) within this object
   */
  public boolean liesWithin(MyRectangle r) {
    return getCircle().liesWithin(r);
  }

  public String getKey() {
    return minefieldhashprefix + getId();
  }

  public String toString() {
    return toString(false);
  }

  public String toString(boolean showXY) {
    return (isweb==1?"Web-":"") + "Minefield " + getId() +
      (showXY?"@("+x+","+y+")":" ") +
      " o"+getOwner() + " with "+units+" mines (r="+getRadius()+")";
  }
}
